############
#
# Copyright (c) 2022 MIT CSAIL and Joseph DelPreto
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
# IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# See https://action-net.csail.mit.edu for more usage information.
# Created 2021-2022 for the MIT ActionNet project by Joseph DelPreto [https://josephdelpreto.com].
#
############

import math
import numpy as np


def wrap_to_180(angle_deg):
    # reduce the angle
    angle_deg = angle_deg % 360

    # force it to be the positive remainder, so that 0 <= angle < 360
    angle_deg = (angle_deg + 360) % 360

    # force into the minimum absolute value residue class, so that -180 < angle <= 180
    if angle_deg > 180:
        angle_deg = angle_deg - 360
    return angle_deg


def euler_from_quaternion(w, x, y, z):
    t0 = +2.0 * (w * x + y * z)
    t1 = +1.0 - 2.0 * (x * x + y * y)
    roll_x_rad = math.atan2(t0, t1)

    t2 = +2.0 * (w * y - z * x)
    t2 = +1.0 if t2 > +1.0 else t2
    t2 = -1.0 if t2 < -1.0 else t2
    pitch_y_rad = math.asin(t2)

    t3 = +2.0 * (w * z + x * y)
    t4 = +1.0 - 2.0 * (y * y + z * z)
    yaw_z_rad = math.atan2(t3, t4)

    eulers_rad = np.array([roll_x_rad, pitch_y_rad, yaw_z_rad])
    return eulers_rad * 180 / np.pi



############
#
# Copyright (c) 2022 MIT CSAIL and Joseph DelPreto
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
# IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# See https://action-net.csail.mit.edu for more usage information.
# Created 2021-2022 for the MIT ActionNet project by Joseph DelPreto [https://josephdelpreto.com].
#
############

from collections import OrderedDict


# Rename a key of a dictionary.
# If it's an OrderedDict, keep the order the same.
def rename_dict_key(d: dict | OrderedDict, old_key, new_key) -> dict | OrderedDict:
    # Check if the rename operation makes sense.
    if old_key not in d:
        raise KeyError(
            "Cannot rename key %s -> %s since the old key does not exist"
            % (old_key, new_key)
        )
    if old_key == new_key:
        return d
    if new_key in d:
        raise AssertionError(
            "Cannot rename key %s -> %s since new key already exists"
            % (old_key, new_key)
        )

    # If it's not an ordered dictionary, simply pop and reassign the value.
    if not isinstance(d, OrderedDict):
        d[new_key] = d.pop(old_key)
        return d

    # Rename the key while preserving key order.
    keys = list(d.keys())
    keys[keys.index(old_key)] = new_key
    return OrderedDict(zip(keys, d.values()))


# Cast all values in a (possibly nested) dictionary to strings.
# Will remove key-value pairs for values that cannot be easily converted to a string.
# If preserve_nested_dicts is True, will preserve nested structure but recursively convert their values.
#   Otherwise, will simply stringify the whole nested dictionary.
def convert_dict_values_to_str(
    d: dict | OrderedDict, preserve_nested_dicts: bool = True
) -> dict | OrderedDict:
    # Create a new dictionary that will be populated.
    if isinstance(d, OrderedDict):
        d_converted = OrderedDict()
    else:
        d_converted = {}

    for key, value in d.items():
        # Recurse if the value is a dictionary.
        if isinstance(value, dict) and preserve_nested_dicts:
            d_converted[key] = convert_dict_values_to_str(
                value, preserve_nested_dicts=preserve_nested_dicts
            )
        else:
            # Add the item to the new dictionary if its value is convertible to a string.
            try:
                d_converted[key] = str(value)
            except:
                pass
    return d_converted


# Flatten a dictionary.
# Will bring items of nested dictionaries up to the root level.
# Keys from nested dictionaries will have the parent key prepended to it.
def flatten_dict(d: dict | OrderedDict) -> dict | OrderedDict:
    d_flattened_items = _get_flattened_dict_items(d)
    if isinstance(d, OrderedDict):
        return OrderedDict(d_flattened_items)
    else:
        return dict(d_flattened_items)


# Worker method for the above, which will return a list of items for the flattened dictionary.
def _get_flattened_dict_items(
    d: dict | OrderedDict, parent_key: str | None = None, parent_key_joiner: str = "|"
) -> list:
    d_items = []
    for key, value in d.items():
        if parent_key is not None:
            key = "%s%s%s" % (parent_key, parent_key_joiner, key)
        if isinstance(value, dict):
            d_items.extend(
                _get_flattened_dict_items(
                    value, parent_key=key, parent_key_joiner=parent_key_joiner
                )
            )
        else:
            d_items.append((key, value))
    return d_items



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

from abc import abstractmethod

from hermes.base.nodes.node_interface import NodeInterface
from hermes.base.stream import Stream


class ProducerInterface(NodeInterface):
    """Interface for the Producer Node component."""

    @classmethod
    @abstractmethod
    def create_stream(cls, stream_spec: dict) -> Stream:
        """Instantiate Stream datastructure object specific to this Pipeline.

        Should also be a class method to create Stream objects on consumers.

        Args:
            stream_spec (dict): Mapping of corresponding Stream object parameters to user-defined configuration values.

        Returns:
            Stream: Datastructure object of the corresponding Node, configured according to the user-provided specification.
        """
        pass

    @abstractmethod
    def _ping_device(self) -> None:
        """Device-specific procedure for round-trip time estimation.

        Concrete implementation of Producer must override the method if required to measure transmission delay
          for realtime/post-processing alignment of modalities that don't support system clock sync.
        """
        pass

    @abstractmethod
    def _connect(self) -> bool:
        """Connect to device via its corresponding backend.

        Returns:
            bool: Whether connection to the device succeeded.
        """
        pass

    @abstractmethod
    def _keep_samples(self) -> None:
        """Node-specific externally triggered function to start keeping in memory streamed data."""
        pass

    @abstractmethod
    def _process_data(self) -> None:
        """Main iteration loop logic for the Node during its running phase.

        Acquire data from your sensor as desired, and for each timestep.
        SDK thread pushes data into shared memory space, this thread pulls data and does all the processing,
        ensuring that lost packets are responsibility of the slow consumer.
        """
        pass

    @abstractmethod
    def _stop_new_data(self) -> None:
        """Stop sampling data, continue sending already captured until none is left."""
        pass



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

import time
from typing import Callable

from hermes.utils.sensor_utils import estimate_transmission_delay
from hermes.utils.time_utils import get_time


class DelayEstimator:
    """Functional callable class for periodic device-specific propagation delay estimation."""

    def __init__(self, sample_period_s: float):
        """Constructor of the DelayEstimator component for propagation delay estimation.

        Args:
            sample_period_s (float): Duration of periods over which to estimate propagation delay.
        """
        self._sample_period_s = sample_period_s
        self._is_continue = True

    def __call__(self, ping_fn: Callable, publish_fn: Callable):
        """Callable that runs periodic propagation delay estimation.

        Uses user-passed estimation and callback functions until termination.

        Args:
            ping_fn (Callable): Propagation delay estimation function pointer.
            publish_fn (Callable): Callback function pointer.
        """
        while self._is_continue:
            delay_s: float = estimate_transmission_delay(ping_fn=ping_fn)
            time_s = get_time()
            publish_fn(time_s, delay_s)
            time.sleep(self._sample_period_s)

    def cleanup(self):
        """Method for external trigger to terminate the delay estimator."""
        self._is_continue = False



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

from abc import abstractmethod
import zmq

from hermes.utils.time_utils import init_time
from hermes.utils.zmq_utils import (
    DNS_LOCALHOST,
    PORT_KILL,
    PORT_KILL_BTN,
    PORT_SYNC_HOST,
    TOPIC_KILL,
)

from hermes.base.nodes.node_interface import NodeInterface
from hermes.base.nodes.node_states import AbstractNodeState, StartState


class Node(NodeInterface):
    """An abstract class with common functionality for concrete Nodes."""

    def __init__(
        self,
        ref_time: float,
        host_ip: str = DNS_LOCALHOST,
        port_sync: str = PORT_SYNC_HOST,
        port_killsig: str = PORT_KILL,
    ) -> None:
        """Constructor of the Node parent class.

        Args:
            ref_time (float): Reference time of the local Broker w.r.t which to align all Nodes.
            host_ip (str, optional): IP address of the local master Broker. Defaults to `DNS_LOCALHOST`.
            port_sync (str, optional): Local port to listen to for local master Broker's startup coordination. Defaults to `PORT_SYNC_HOST`.
            port_killsig (str, optional): Local port to listen to for local master Broker's termination signal. Defaults to `PORT_KILL`.
        """
        self._host_ip = host_ip
        self._port_sync = port_sync
        self._port_killsig = port_killsig
        self.__is_done = False
        self._ref_time_s = ref_time
        init_time(ref_time=ref_time)

        self._state = StartState(self)

        self._ctx: zmq.Context = zmq.Context.instance()
        self._poller: zmq.Poller = zmq.Poller()

    @property
    def _is_done(self) -> bool:
        return self.__is_done

    @_is_done.setter
    def _is_done(self, done: bool) -> None:
        self.__is_done = done

    def __call__(self):
        """Node objects are callable to start the FSM as entry-point."""
        while self._state.is_continue():
            self._state.run()
        self._cleanup()
        print("%s exited, goodbye <3" % self._log_source_tag(), flush=True)

    def _set_state(self, state: AbstractNodeState) -> None:
        self._state = state

    @abstractmethod
    def _initialize(self):
        # Socket to receive kill signal
        self._killsig: zmq.SyncSocket = self._ctx.socket(zmq.SUB)
        self._killsig.connect("tcp://%s:%s" % (DNS_LOCALHOST, self._port_killsig))
        topics = [TOPIC_KILL]
        for topic in topics:
            self._killsig.subscribe(topic)

        # Socket to indicate to broker that the subscriber is ready
        self._sync: zmq.SyncSocket = self._ctx.socket(zmq.REQ)
        self._sync.connect("tcp://%s:%s" % (self._host_ip, self._port_sync))
        # Socket to indicate to broker that the Node caught interrupt signal
        self._babykillsig: zmq.SyncSocket = self._ctx.socket(zmq.REQ)
        self._babykillsig.connect("tcp://%s:%s" % (DNS_LOCALHOST, PORT_KILL_BTN))

    def _get_sync_socket(self) -> zmq.SyncSocket:
        return self._sync

    def _activate_kill_poller(self) -> None:
        self._poller.register(self._killsig, zmq.POLLIN)

    @abstractmethod
    def _activate_data_poller(self) -> None:
        pass

    def _deactivate_kill_poller(self) -> None:
        print("%s received KILL signal" % self._log_source_tag(), flush=True)
        # self._killsig.recv_multipart()
        self._poller.unregister(self._killsig)

    def _send_kill_to_broker(self):
        self._babykillsig.send_string(TOPIC_KILL)

    def _poll(self) -> tuple[list[zmq.SyncSocket], list[int]]:
        return tuple(zip(*(self._poller.poll())))

    @abstractmethod
    def _on_poll(self, poll_res: tuple[list[zmq.SyncSocket], list[int]]) -> None:
        if self._killsig in poll_res[0]:
            self._state.kill()

    @abstractmethod
    def _trigger_stop(self) -> None:
        pass

    @abstractmethod
    def _cleanup(self):
        """Release of generic Node resources, must be done after releasing higher-level resources."""
        self._killsig.close()
        self._babykillsig.close()
        self._sync.close()
        # Destroy ZeroMQ context.
        self._ctx.term()



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

# ZeroMQ topics and message strings
TOPIC_KILL = "KILL"
CMD_HELLO = "HELLO"
CMD_ACK = "ACK"
CMD_START_TIME = "START_TIME"
CMD_GO = "GO"
CMD_END = "END"
CMD_EXIT = "EXIT?"
CMD_BYE = "BYE"
MSG_ON = "ON"
MSG_OFF = "OFF"
MSG_OK = "OK"

# Ports used for ZeroMQ by our system
PORT_BACKEND = "42069"
PORT_FRONTEND = "42070"
PORT_SYNC_HOST = "42071"
PORT_SYNC_REMOTE = "42072"
PORT_KILL = "42066"
PORT_KILL_BTN = "42065"
PORT_PAUSE = "42067"

# Ports of connected devices/sensors
PORT_MOTICON = "8888"  # defined by the Moticon desktop app, putting data at the loopback address for listening
PORT_PROSTHESIS = "51702"  # defined by LabView code of VUB
PORT_GUI = "8005"
PORT_EYE = "50020"
PORT_VICON = "801"  # defined by Vicon software
PORT_MVN = "9763"  # defined by MVN Analyze

# IP addresses of devices on the network used by our system
DNS_LOCALHOST = "localhost"
IP_LOOPBACK = "127.0.0.1"

# TODO: remove hardcoding
IP_TERMINAL = "192.168.0.100"
IP_PROSTHESIS = "192.168.0.101"
IP_BACKPACK = "192.168.0.103"
IP_VICON = "192.168.0.104"



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

from collections import namedtuple
from dataclasses import dataclass
from typing import Optional, TypeAlias, Any, Deque, Iterable, Mapping, Dict
from enum import Enum
from threading import Lock
import zmq


NewDataDict: TypeAlias = Dict[str, Dict[str, Any]]
DataFifo: TypeAlias = Deque[Any]
DataFifoDict: TypeAlias = Dict[str, Dict[str, DataFifo]]
StreamInfoDict: TypeAlias = Dict[str, Dict[str, Dict[str, Any]]]
DeviceLockDict: TypeAlias = Dict[str, Lock]
ExtraDataInfoDict: TypeAlias = Dict[str, Dict[str, Any]]
VideoFormatTuple = namedtuple("VideoFormatTuple", ("format", "color"))
AudioFormatTuple = namedtuple("AudioFormatTuple", ("format", "color"))
ZMQResult: TypeAlias = Iterable[tuple[zmq.SyncSocket, int]]


@dataclass
class VideoCodec:
    """Object specifying video codec options for FFmpeg."""

    codec_name: str
    pix_format: str
    num_cpu: int = 1
    input_options: Mapping = None
    output_options: Mapping = None


@dataclass
class AudioCodec:
    """Object specifying audio codec options for FFmpeg."""

    codec_name: str
    pix_format: str
    num_cpu: int = 1
    input_options: Mapping = None
    output_options: Mapping = None


@dataclass
class LoggingSpec:
    """Object specifying data storage options.

    Args:
        log_dir (str): Path to the directory on disk to flush data to.
        experiment (dict[str, str]): Nested setup definition of Nodes across distributed hosts.
        log_time_s (float): Start time of saving data.
        ref_time_s (float): Reference time of the Broker to align all Nodes to.
        stream_period_s (float, optional): Duration of periods over which to flush streamed accumulated data from memory to disk. Defaults to `30.0`.
        is_quiet (bool): Whether to print FFmpeg stats to the terminal. Defaults to `False`.
        stream_hdf5 (bool, optional): Whether to stream data into HDF5 files. Defaults to `False`.
        stream_video (bool, optional): Whether to stream video data into MP4/MKV files. Defaults to `False`.
        stream_csv (bool, optional): Whether to stream data into CSV files. Defaults to `False`.
        stream_audio (bool, optional): Whether to stream audio data into MP3/WAV files. Defaults to `False`.
        dump_csv (bool, optional): Weather to dump in-memory recorded data in CSV files. Defaults to `False`.
        dump_hdf5 (bool, optional): Weather to dump in-memory recorded data in HDF5 files. Defaults to `False`.
        dump_video (bool, optional): Weather to dump in-memory recorded video data in MP4/MKV files. Defaults to `False`.
        dump_audio (bool, optional): Weather to dump in-memory recorded audio data in MP3/WAV files. Defaults to `False`.
        video_codec (VideoCodec, optional): Definition of the video codec to use for FFmpeg. Defaults to `None`.
        audio_codec (AudioCodec, optional): Definition of the audio codec to use for FFmpeg. Defaults to `None`.
    """

    log_dir: str
    experiment: dict[str, str]
    log_time_s: float
    ref_time_s: float
    stream_period_s: Optional[float] = 30.0
    is_quiet: Optional[bool] = False
    stream_hdf5: Optional[bool] = False
    stream_video: Optional[bool] = False
    stream_csv: Optional[bool] = False
    stream_audio: Optional[bool] = False
    dump_hdf5: Optional[bool] = False
    dump_video: Optional[bool] = False
    dump_csv: Optional[bool] = False
    dump_audio: Optional[bool] = False
    video_codec: Optional[VideoCodec] = None
    audio_codec: Optional[AudioCodec] = None


class VideoFormatEnum(Enum):
    """Video format enumeration for supported FFmpeg video formats.

    Must be a tuple of (<FFmpeg write format>, <Color format>), where:
        write format is one of: `ffmpeg -formats`
        pixel color is one of: `ffmpeg -pix_fmts`
    """

    BGR = VideoFormatTuple("rawvideo", "bgr24")
    YUV = VideoFormatTuple("rawvideo", "yuv420p")
    JPEG = VideoFormatTuple("image2pipe", "yuv420p")
    MJPEG = VideoFormatTuple("jpeg_pipe", "yuv420p")
    BAYER_RG8 = VideoFormatTuple("rawvideo", "bayer_rggb8")


class AudioFormatEnum(Enum):
    """Audio format enumeration for supported FFmpeg video formats.

    TODO:
    Must be a tuple of (<FFmpeg write format>, ...), where:
        write format is one of: `ffmpeg -formats`
    """

    BGR = AudioFormatTuple("rawvideo", "bgr24")
    YUV = AudioFormatTuple("rawvideo", "yuv420p")
    JPEG = AudioFormatTuple("image2pipe", "yuv420p")
    BAYER_RG8 = AudioFormatTuple("rawvideo", "bayer_rggb8")



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

from abc import ABC, abstractmethod
import queue
from typing import Any, Callable, Iterable
from collections import OrderedDict, deque


class BufferInterface(ABC):
    """Interface for the multi-channel FIFO buffer."""

    @abstractmethod
    def plop(self, key: str, data: dict) -> None:
        """Asynchronously adds an element to the specified channel of the buffer.

        Args:
            key (str): Unique identifier of the channel to add the data to.
            data (dict): Usecase-specific nested dictionary of data to add.
        """
        pass

    @abstractmethod
    def yeet(self) -> Any:
        """Synchronously retrieves the oldest set of samples from all channels of the buffer.

        Returns:
            Any: Multi-channel vector of the oldest sample.
        """
        pass


class NonOverflowingCounterConverter:
    """A counter value converter from overflowing fixed range to non-overflowing values, starting at 0 for the first received sample.

    Converts overflowing monotonically increasing counter from a sensor into a non-overflowing value,
    starting counting from 0, regardless of device's actual onboard counter.
    """

    def __init__(self, keys: Iterable[Any], num_bits_counter: int):
        """Constructor of the NonOverflowingCounterConverter.

        Args:
            keys (Iterable[Any]): Set of uniquely identifying channel keys.
            num_bits_counter (int): The fixed-width of the counter generating the data.
        """
        self._counter_limit = 2**num_bits_counter
        self._convert_fn: Callable[[Any, int], int | None] = self._foo
        self._start_counters: OrderedDict[Any, int | None] = OrderedDict(
            [(k, None) for k in keys]
        )
        self._previous_counters: OrderedDict[Any, int | None] = OrderedDict(
            [(k, None) for k in keys]
        )
        self._counters: OrderedDict[Any, int | None] = OrderedDict(
            [(k, None) for k in keys]
        )
        self._start_counter = None

    def _foo(self, key: Any, counter: int) -> int | None:
        """Startup multi-channel function that converts correlated over-flowing counter values from asynchronous sensors to a non-overflowing counter.

        Switches to the continuous function when all channels have provided a starting sample.
        Else, reuses this function with branch conditional logic.
        The branched version is not ideal for continuous use, because other conditions never happen after the first iteration of the module on the channel.

        Args:
            key (Any): The unique identifier of the channel.
            counter (int): Monotonically increasing overflowing integer from a sensor.

        Returns:
            int | None: Converted counter value starting at 0 counter value.
        """
        if self._start_counter is None:
            self._start_counter = counter
            self._start_counters[key] = counter
            self._previous_counters[key] = counter
            self._counters[key] = 0
        elif self._previous_counters[key] is None:
            self._start_counters[key] = counter
            self._previous_counters[key] = counter
            dc = (counter - self._start_counter) % self._counter_limit
            self._counters[key] = dc
        else:
            self._bar(key=key, counter=counter)
        if all([counter is not None for counter in self._start_counters.values()]):
            self._convert_fn = self._bar
        return self._counters[key]

    def _bar(self, key: Any, counter: int) -> int | None:
        """Optimized counter converter function for continuous steady-state operation.

        Previous counters are guaranteed to be non-0 after this function is activated.

        Args:
            key (Any): The unique identifier of the channel.
            counter (int): Monotonically increasing overflowing integer from a sensor.

        Returns:
            int | None: Converted counter value starting at 0 counter value.
        """
        dc = (counter - self._previous_counters[key]) % self._counter_limit  # type: ignore
        self._previous_counters[key] = counter
        self._counters[key] += dc
        return self._counters[key]


class TimestampToCounterConverter:
    """A counter value converter from overflowing fixed range timestamp to non-overflowing counter values, starting at 0 for the first received sample.

    Converts overflowing monotonically increasing timestamp of a certain sample rate from a sensor into a non-overflowing counter value,
    starting counting from 0, regardless of device's actual onboard timestamp.
    """

    def __init__(self, keys: Iterable[Any], sampling_period: int, counter_limit: int):
        """Constructor of the TimestampToCounterConverter.

        Args:
            keys (Iterable[Any]): Set of uniquely identifying channel keys.
            sampling_period (int): Sampling period in the same units as timestamp limit and timestamps.
            counter_limit (int): The upper counting limit of the sensor's timestamp.
        """
        self._sampling_period = sampling_period
        self._timestamp_limit: int = counter_limit
        self._counter_from_timestamp_fn: Callable[
            [Any, int | float], int | float | None
        ] = self._foo
        self._first_timestamps: OrderedDict[Any, int | float | None] = OrderedDict(
            [(k, None) for k in keys]
        )
        self._previous_timestamps: OrderedDict[Any, int | float | None] = OrderedDict(
            [(k, None) for k in keys]
        )
        self._counters: OrderedDict[Any, int | None] = OrderedDict(
            [(k, None) for k in keys]
        )

    def _foo(self, key: Any, timestamp: int | float) -> int | float | None:
        """Startup multi-channel function that converts correlated over-flowing timestamp values from asynchronous sensors to a non-overflowing counter.

        Sets the start time according to the first received packet and switches to the monotone calculation routine after.
        Has some tolerance to temporally skewed samples, when the skew is less than hald a sampling period.
        If the channel sample is the first in the overall buffer, will use it as reference starting point onward.
        Will return 0 start counter at the end of the function.

        If it's not the very first packet, but first reading for this device, records if the capture was during or after the start reference.
        If the measurement taken during or after the reference measurement and no chance for overflow, will return 0 start counter at the end of the function.
        If the measurement taken after the overflow of the on-sensor clock and effectively after the reference measurement,
        will return 0 start counter at the end of the function.
        Will discard the sample as stale to ensure alignment otherwise.

        Args:
            key (Any): The unique identifier of the channel.
            timestamp (int): Monotonically increasing overflowing integer from a sensor.

        Returns:
            int | None: Converted counter value starting at 0 counter value.
        """
        if not any([v is not None for v in self._previous_timestamps.values()]):
            self._start_time = timestamp
            self._first_timestamps[key] = timestamp
            self._previous_timestamps[key] = timestamp
            self._counters[key] = 0
        elif self._previous_timestamps[key] is None:
            if timestamp >= self._start_time:
                self._first_timestamps[key] = timestamp
                self._previous_timestamps[key] = timestamp
                self._counters[key] = round(
                    ((timestamp - self._start_time) % self._timestamp_limit)
                    / self._sampling_period
                )
            elif ((timestamp - self._start_time) % self._timestamp_limit) < (
                self._start_time - timestamp
            ):
                self._first_timestamps[key] = timestamp
                self._previous_timestamps[key] = timestamp
                self._counters[key] = round(
                    ((timestamp - self._start_time) % self._timestamp_limit)
                    / self._sampling_period
                )
            else:
                return None
        else:
            self._bar(key=key, timestamp=timestamp)
        if all([v is not None for v in self._previous_timestamps.values()]):
            self._counter_from_timestamp_fn = self._bar
        return self._counters[key]

    def _bar(self, key: Any, timestamp: int | float) -> int | float | None:
        """Optimized counter converter function for continuous steady-state operation.

        Measures the change in time between 2 measurements w.r.t. sensor device time and the max value before overlow.
        dt > 0 always thanks to modulo, even if sensor on-board clock overflows.
        Converts to the number of sample periods in the measured time delta window, allowing for slight skew.
        Rolling correlation using sample rate, previous and current time is more accurate than averaging over whole timelife.

        Args:
            key (Any): The unique identifier of the channel.
            timestamp (int): Monotonically increasing overflowing integer from a sensor.

        Returns:
            int | None: Converted counter value starting at 0 counter value.
        """
        delta_ticks = (timestamp - self._previous_timestamps[key]) % self._timestamp_limit  # type: ignore
        self._previous_timestamps[key] = timestamp
        delta_counter = round(delta_ticks / self._sampling_period)
        self._counters[key] += delta_counter  # type: ignore
        return self._counters[key]


class AlignedFifoBuffer(BufferInterface):
    """Multichannel first-in first-out buffer that aligns asynchronous temporally-lossy samples across channels.

    Receives asynchronous samples for each channel, aligns them, and returns to the user an aligned snapshot across all channels.
    Allows yeeting from buffer if some keys have been empty for a while (disconnection or out of range),
    while others continue producing.

    By default uses dynamically-growing Deque for the buffer, approprate for the sample rate of IMUs.
    `maxlen` offers possibility to turn into a fixed-length ring buffer, to avoid unnecessary memory allocations for higher performance,
    at the cost of lost data in case of slow consumers.

    Updates only on yeet to discard stale sample that arrived too late.
    Adds counter into the data payload to retreive on the reader. (Useful for time->counter converted buffer).
    If the snapshot had not been read, even if the measurement is stale (arrived later than specified), still adds it to the buffer.
    Empty pads if some intermediate timesteps did not recieve a packet for a specific key.
    If buffer contents are valid, moves snapshot into the output Queue.
    Update the frame counter to keep track of removed data to discard stale late arrivals.
    """

    def __init__(
        self, keys: Iterable, timesteps_before_stale: int, maxlen: int | None = None
    ):
        """Constructor of the AlignedFifoBuffer.

        Args:
            keys (Iterable): Set of uniquely identifying channel keys.
            timesteps_before_stale (int): The number of samples in other channels after which a missing sample in a channel is marked missing.
            maxlen (int | None): Fixed length of preallocated ring buffer. Defaults to None.
        """
        self._buffer = OrderedDict([(k, deque(maxlen=maxlen)) for k in keys])
        self._output_queue: queue.Queue[dict] = queue.Queue()
        self._counter_snapshot = 0
        self._timesteps_before_stale = timesteps_before_stale

    def plop(self, key: str, data: dict, **kwargs):
        counter: int = kwargs["counter"]
        data["counter"] = counter
        if counter >= self._counter_snapshot:
            while len(self._buffer[key]) < (counter - self._counter_snapshot):
                self._buffer[key].append(None)
            self._buffer[key].append(data)
        else:
            print("%d packet of %s arrived too late." % (counter, key), flush=True)
        is_every_key_has_data = all([len(buf) for buf in self._buffer.values()])
        is_some_key_exceeds_stale_period = any(
            [len(buf) >= self._timesteps_before_stale for buf in self._buffer.values()]
        )
        is_some_key_empty = any([not len(buf) for buf in self._buffer.values()])
        if is_every_key_has_data:
            oldest_packet = {k: buf.popleft() for k, buf in self._buffer.items()}
            self._put_output_queue(oldest_packet)
        elif is_some_key_exceeds_stale_period and is_some_key_empty:
            oldest_packet = {
                k: (buf.popleft() if len(buf) else None)
                for k, buf in self._buffer.items()
            }
            self._put_output_queue(oldest_packet)

    def _put_output_queue(self, packet: dict) -> None:
        """Places a ready to consume complete snapshot onto the output queue.

        Args:
            packet (dict): Temporally aligned snapshot mapping unique channel-identifying keys to the corresponding sample.
        """
        self._counter_snapshot += 1
        self._output_queue.put(packet)

    def flush(self) -> None:
        """Allow to evict all present data because no new samples will be captured."""
        while is_any_key_not_empty := any([len(buf) for buf in self._buffer.values()]):
            oldest_packet = {
                k: (buf.popleft() if len(buf) else None)
                for k, buf in self._buffer.items()
            }
            self._put_output_queue(oldest_packet)

    def yeet(self, timeout: float = 10.0) -> dict | None:
        """Attempts to synchronously retrieve the oldest set of samples from all channels of the buffer with a timeout.

        Args:
            timeout (float, optional): How long to wait for new snapshot. Defaults to `10.0`.

        Returns:
            dict | None: Multi-channel vector of the oldest sample or None if no new data became available until timeout.
        """
        try:
            return self._output_queue.get(timeout=timeout)
        except queue.Empty:
            print("Timed out on no more snapshots in the output Queue.")
            return None


class TimestampAlignedFifoBuffer(AlignedFifoBuffer):
    """Multichannel first-in first-out buffer that aligns asynchronous temporally-lossy samples across channels by supplied correlated timestamps.

    Allows yeeting from buffer if some keys have been empty for a while, while others continue producing.
    """

    def __init__(
        self,
        keys: Iterable,
        timesteps_before_stale: int,
        sampling_period: int,
        counter_limit: int,
        maxlen: int | None = None,
    ):
        """Constructor of the TimestampAlignedFifoBuffer.

        Args:
            keys (Iterable): Set of uniquely identifying channel keys.
            timesteps_before_stale (int): The number of samples in other channels after which a missing sample in a channel is marked missing.
            sampling_period (int): Sampling period in the same units as timestamp limit and timestamps.
            counter_limit (int): The upper counting limit of the sensor's timestamp.
            maxlen (int | None): Fixed length of preallocated ring buffer. Defaults to `None`.
        """
        super().__init__(
            keys=keys, timesteps_before_stale=timesteps_before_stale, maxlen=maxlen
        )
        self._converter = TimestampToCounterConverter(
            keys=keys, sampling_period=sampling_period, counter_limit=counter_limit
        )

    def plop(self, key: str, data: dict, **kwargs) -> None:
        timestamp: float = kwargs["timestamp"]
        counter = self._converter._counter_from_timestamp_fn(key, timestamp)
        if counter is not None:
            super().plop(key=key, data=data, counter=counter)


class NonOverflowingCounterAlignedFifoBuffer(AlignedFifoBuffer):
    """Multichannel first-in first-out buffer that aligns asynchronous temporally-lossy samples across channels by supplied correlated overflowing counter."""

    def __init__(
        self,
        keys: Iterable,
        timesteps_before_stale: int,
        num_bits_timestamp: int,
        maxlen: int | None = None,
    ):
        """Constructor of the NonOverflowingCounterAlignedFifoBuffer.

        Args:
            keys (Iterable): Set of uniquely identifying channel keys.
            timesteps_before_stale (int): The number of samples in other channels after which a missing sample in a channel is marked missing.
            num_bits_timestamp (int): The fixed-width of the counter generating the data.
            maxlen (int | None): Fixed length of preallocated ring buffer. Defaults to `None`.
        """
        super().__init__(
            keys=keys, timesteps_before_stale=timesteps_before_stale, maxlen=maxlen
        )
        self._converter = NonOverflowingCounterConverter(
            keys=keys, num_bits_counter=num_bits_timestamp
        )

    def plop(self, key: str, data: dict, **kwargs) -> None:
        raw_counter: int = kwargs["counter"]
        counter = self._converter._convert_fn(key, raw_counter)
        if counter is not None:
            super().plop(key=key, data=data, counter=counter)



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

import time
from typing import Callable
import numpy as np


def estimate_transmission_delay(ping_fn: Callable, num_samples: int = 100) -> float:
    """Estimates the mean transmission delay of the provided "ping" function.

    Used with the custom provided `ping_fn` to measure the one-way delay to use for
    offsetting sensor data streams from the `toa_s` to obtain true sample time.

    Args:
        ping_fn (Callable): User's function that wraps "ping" like functionality to the specific device.
        num_samples (int, optional): Number of round-trip transmission to perform to average over. Defaults to `100`.

    Returns:
        float: Estimated mean delay for samples from the device.
    """
    transmit_delays_s: list[float] = []
    for i in range(num_samples):
        local_time_before = time.time()
        ping_fn()
        local_time_after = time.time()
        # Assume symmetric delays.
        transmit_delays_s.append((local_time_after - local_time_before) / 2.0)
    # TODO: remove outliers before averaging.
    delays = np.array(transmit_delays_s)
    return float(np.mean(delays))



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

from collections import OrderedDict
from io import TextIOWrapper
from subprocess import Popen
import os
import time
import asyncio
import concurrent.futures
import h5py
import numpy as np
from typing import Any, Iterator

try:
    import ffmpeg
except ImportError as e:
    print(
        e,
        "\nFFmpeg not installed, will crash if you configure streaming of video/audio.",
        flush=True,
    )

from hermes.base.stream import Stream
from hermes.base.storage.storage_interface import StorageInterface
from hermes.base.storage.storage_states import AbstractStorageState, StartState
from hermes.utils.time_utils import init_time, get_time, get_time_str
from hermes.utils.dict_utils import convert_dict_values_to_str
from hermes.utils.types import LoggingSpec


class Storage(StorageInterface):
    """Manages IO operations of all stream data.

    Flushes data periodically for continuous operation and clears from memory
    to reduce RAM usage, or all at once if user guarantees enough memory.
    In continuous mode, will flush leftover data once the program is stopped.

    Logs video and audio data with FFmpeg to MKV/MP4 and MP3, respectively.
    Logs all other sensor data in a single hierarchical HDF5 file.
    CSV format is also supported, but discouraged -> creates file per sub-stream.

    If sub-stream elements contain a burst of samples of sample_size,
    will automatically unroll it.

    Will fail if no FFmpeg is installed on the system.
    """

    def __init__(self, log_tag: str, spec: LoggingSpec):
        """Constructor of the Storage component responsible for all IO.

        Args:
            log_tag (str): Filename prefix.
            spec (LoggingSpec): Specifies what and how to store to disk.
        """
        self._log_tag = log_tag
        self._spec = spec

        # Initialize variables that will guide the thread that will do stream/dump logging of data available in the Stream objects.
        #   Main thread will listen to the sockets and put files to the Stream objects.
        self._is_streaming: bool  # whether periodic writing is active
        self._is_flush: bool  # whether remaining data at the end should now be flushed
        self._is_finished: bool  # whether the logging loop is finished and all data was flushed

        # Initialize the logging writers.
        self._thread_pool: concurrent.futures.ThreadPoolExecutor
        self._hdf5_file: h5py.File | None = None
        self._video_writers: list[tuple[Popen, str, str, str]] = []
        self._audio_writers: list[tuple[Popen, str, str, str]] = []
        self._csv_writers: list[tuple[TextIOWrapper, str, str, str]] = []
        self._csv_writer_metadata: TextIOWrapper | None = None

        # Create the log directory if needed.
        if self._is_to_stream() or self._is_to_dump():
            os.makedirs(self._spec.log_dir, exist_ok=True)

    def __call__(self, streams: OrderedDict[str, Stream]) -> None:
        """Callable that runs main FSM loop.

        Runs continuously, ignoring Ctrl+C interrupt, until owner Node triggers an exit.

        Args:
            streams (OrderedDict[str, Stream]): Reference to the Stream objects to flush to disk.
        """
        init_time(ref_time=self._spec.ref_time_s)
        self._state = StartState(self, streams)
        while self._state.is_continue():
            self._state.run()
        print("%s Logger safely exited." % self._log_tag, flush=True)

    def cleanup(self) -> None:
        """Stop stream-logging and wait for it to finish.

        Will stop stream-logging, if it is active.
        Will trigger data dump, if configured.
        Node pushing data to the Stream should stop adding new data before cleaning up Logger.
        """
        self._stop_stream_logging()
        self._log_stop_time_s = get_time()

    ############################
    ###### FSM OPERATIONS ######
    ############################
    def _initialize(self, streams: OrderedDict[str, Stream]) -> None:
        self._hdf5_log_length_increment = 10000
        self._streams = streams
        self._timesteps_before_solidified: OrderedDict[
            str, OrderedDict[str, OrderedDict[str, int]]
        ] = OrderedDict()
        self._next_data_indices_hdf5: OrderedDict[
            str, OrderedDict[str, OrderedDict[str, int]]
        ] = OrderedDict()
        for tag in streams.keys():
            # Initialize a record of what indices have been logged,
            #  and how many timesteps to stay behind of the most recent step (if needed).
            self._timesteps_before_solidified.setdefault(tag, OrderedDict())
            # Each time an HDF5 dataset reaches its limit,
            #  its size will be increased by the following amount.
            self._next_data_indices_hdf5.setdefault(tag, OrderedDict())

    def _set_state(self, state: AbstractStorageState) -> None:
        self._state = state

    def _is_to_stream(self) -> bool:
        return (
            self._spec.stream_csv
            or self._spec.stream_hdf5
            or self._spec.stream_video
            or self._spec.stream_audio
        )

    def _is_to_dump(self) -> bool:
        return (
            self._spec.dump_csv
            or self._spec.dump_hdf5
            or self._spec.dump_video
            or self._spec.dump_audio
        )

    def _start_stream_logging(self) -> None:
        num_workers: int = 0
        if self._spec.stream_csv:
            num_workers += self._init_files_csv()
        if self._spec.stream_hdf5:
            num_workers += self._init_files_hdf5()
        if self._spec.stream_video:
            num_workers += self._init_files_video()
        if self._spec.stream_audio:
            num_workers += self._init_files_audio()
        self._init_log_indices()
        self._thread_pool = concurrent.futures.ThreadPoolExecutor(
            max_workers=num_workers
        )
        self._is_streaming = True
        self._is_flush = False
        self._is_finished = False

    def _stop_stream_logging(self) -> None:
        self._is_streaming = False
        self._is_flush = True

    def _start_dump_logging(self) -> None:
        num_workers: int = 0
        if self._spec.dump_csv:
            num_workers += self._init_files_csv()
        if self._spec.dump_hdf5:
            num_workers += self._init_files_hdf5()
        if self._spec.dump_video:
            num_workers += self._init_files_video()
        if self._spec.dump_audio:
            num_workers += self._init_files_audio()
        # Log all data.
        # Will basically enable periodic stream-logging,
        #  but will set self._is_flush and self._is_streaming such that
        #  it seems like the experiment ended and just a final flush is required.
        #  This will cause the stream-logging in self._log_data()
        #  to fetch and write any outstanding data, which is all data since
        #  none is written yet.  It will then exit after the single write.
        # Pretend like the dumping options are actually streaming options.
        self._spec.stream_csv = self._spec.dump_csv
        self._spec.stream_hdf5 = self._spec.dump_hdf5
        self._spec.stream_video = self._spec.dump_video
        self._spec.stream_audio = self._spec.dump_audio
        # Clear the is_finished flag in case dump is run after stream so the log loop
        #  can run once to flush all logged data that wasn't stream-logged.
        self._is_finished = False
        # Initialize indices and log all of the data.
        self._init_log_indices()
        self._thread_pool = concurrent.futures.ThreadPoolExecutor(
            max_workers=num_workers
        )

    def _wait_till_flush(self) -> None:
        while not self._is_flush:
            time.sleep(self._spec.stream_period_s)

    def _release_thread_pool(self) -> None:
        self._thread_pool.shutdown()

    #############################
    ###### FILE OPERATIONS ######
    #############################
    def _init_log_indices(self) -> None:
        """Initialize the data indices to fetch for logging

        Will record the next data indices that should be fetched for each stream,
        and the number of timesteps that each streamer needs before data is solidified.
        """
        for streamer_name, stream in self._streams.items():
            for device_name, device_info in stream.get_stream_info_all().items():
                self._timesteps_before_solidified[streamer_name][
                    device_name
                ] = OrderedDict()
                for stream_name, stream_info in device_info.items():
                    self._timesteps_before_solidified[streamer_name][device_name][
                        stream_name
                    ] = stream_info["timesteps_before_solidified"]

    def _init_files_csv(self) -> int:
        """Create and initialize CSV files.

        Will have a separate file for each stream of each device.
        Currently assumes that device names are unique across all streamers.

        Returns:
            int: Number of initialized writers.
        """
        num_writers: int = 0
        for streamer_name, streamer in self._streams.items():
            for device_name, device_info in streamer.get_stream_info_all().items():
                for stream_name, stream_info in device_info.items():
                    # Skip saving video or audio in a CSV.
                    if stream_info["is_video"] or stream_info["is_audio"]:
                        continue
                    filename_csv = "%s_%s_%s.csv" % (
                        self._log_tag,
                        device_name,
                        stream_name,
                    )
                    filepath_csv = os.path.join(self._spec.log_dir, filename_csv)
                    csv_writer = open(filepath_csv, "w")
                    self._csv_writers.append(
                        (csv_writer, streamer_name, device_name, stream_name)
                    )
                    num_writers += 1

        # Open a writer for a CSV metadata file.
        filename_csv = "%s__metadata.csv" % (self._log_tag)
        filepath_csv = os.path.join(self._spec.log_dir, filename_csv)
        self._csv_writer_metadata = open(filepath_csv, "w")

        # Write CSV headers.
        for stream_writer, streamer_name, device_name, stream_name in self._csv_writers:
            # First check if custom header titles have been specified.
            stream_info = self._streams[streamer_name].get_stream_info_all()[
                device_name
            ][stream_name]
            sample_size = stream_info["sample_size"]
            if (
                isinstance(stream_info["data_notes"], dict)
                and Stream.metadata_data_headings_key in stream_info["data_notes"]
            ):
                data_headers = stream_info["data_notes"][
                    Stream.metadata_data_headings_key
                ]
            else:
                # Write a number of data headers based on how many values are in each data sample.
                # Each sample may be a matrix that will be unwrapped into columns,
                #  so label headers as i-j where i is the original matrix row
                #  and j is the original matrix column (and if more than 2D keep adding more).
                data_headers = []
                subs = np.unravel_index(range(0, np.prod(sample_size)), sample_size)
                subs = np.stack(subs).T
                for header_index in range(subs.shape[0]):
                    header = "Data Entry "
                    for sub_index in range(subs.shape[1]):
                        header += "%d-" % subs[header_index, sub_index]
                    header = header.strip("-")
                    data_headers.append(header)
            stream_writer.write(",")
            stream_writer.write(",".join(data_headers))
        return num_writers

    def _init_files_hdf5(self) -> int:
        """Create and initialize a hierarchical HDF5 file.

        Will have a single file for all streams from all devices.
        Currently assumes that device names are unique across all streamers.

        Returns:
            int: Number of initialized writers.
        """
        filename_hdf5 = "%s.hdf5" % self._log_tag
        filepath_hdf5 = os.path.join(self._spec.log_dir, filename_hdf5)
        num_to_append = 0
        while os.path.exists(filepath_hdf5):
            num_to_append += 1
            filename_hdf5 = "%s_%02d.hdf5" % (self._log_tag, num_to_append)
            filepath_hdf5 = os.path.join(self._spec.log_dir, filename_hdf5)
        self._hdf5_file = h5py.File(filepath_hdf5, "w")
        # Create a dataset for each data key of each stream of each device.
        for streamer_name, stream in self._streams.items():
            streamer_group = self._hdf5_file.create_group(streamer_name)
            for device_name, device_info in stream.get_stream_info_all().items():
                device_group = streamer_group.create_group(device_name)
                self._next_data_indices_hdf5[streamer_name][device_name] = OrderedDict()
                for stream_name, stream_info in device_info.items():
                    # Skip saving video and audio in the HDF5.
                    if stream_info["is_video"] or stream_info["is_audio"]:
                        continue
                    self._next_data_indices_hdf5[streamer_name][device_name][
                        stream_name
                    ] = 0
                    # The main data has specifications defined by stream_info.
                    sample_size = stream_info["sample_size"]
                    data_type = stream_info["data_type"]
                    # Create the dataset.
                    device_group.create_dataset(
                        name=stream_name,
                        shape=(self._hdf5_log_length_increment, *sample_size),
                        maxshape=(None, *sample_size),
                        dtype=data_type,
                        chunks=True,
                    )
        return 1

    def _init_files_video(self) -> int:
        """Create and initialize video writers, one for each device.

        Will fail if no FFmpeg installed.

        Raises:
            ValueError: When no supported codec specification was provided in config file.

        Returns:
            int: Number of initialized writers.
        """
        if self._spec.video_codec is None:
            raise ValueError(
                "Must provide video codec specification when streaming video."
            )

        num_writers: int = 0
        for streamer_name, streamer in self._streams.items():
            for device_name, device_info in streamer.get_stream_info_all().items():
                for stream_name, stream_info in device_info.items():
                    # Skip non-video streams.
                    if not stream_info["is_video"]:
                        continue
                    # Create a unique file.
                    filename_base = "%s_%s" % (self._log_tag, device_name)
                    filename_video = "%s.mkv" % (filename_base)
                    filepath_video = os.path.join(self._spec.log_dir, filename_video)
                    num_to_append = 0
                    while os.path.exists(filepath_video):
                        num_to_append += 1
                        filename_video = "%s_%02d.mkv" % (filename_base, num_to_append)
                        filepath_video = os.path.join(
                            self._spec.log_dir, filename_video
                        )
                    # Create a video writer.
                    frame_height = stream_info["sample_size"][0]
                    frame_width = stream_info["sample_size"][1]
                    fps = stream_info["sampling_rate_hz"]
                    input_stream_format: str = stream_info["format"]
                    input_stream_pix_fmt: str = stream_info["color"]
                    metadata_dict = {
                        "metadata:g:%d" % i: "%s=%s" % (k, v)
                        for i, (k, v) in enumerate(
                            [
                                ("title", "/".join(self._spec.experiment.values())),
                                (
                                    "date",
                                    get_time_str(self._spec.log_time_s, "%Y-%m-%d"),
                                ),
                                (
                                    "comment",
                                    "HERMES multi-modal data acquisition system recording",
                                ),
                                *map(
                                    lambda tup: ("X%s" % tup[0], tup[1]),
                                    list(self._spec.experiment.items()),
                                ),
                                ("Xencoder", self._spec.video_codec.codec_name),
                                ("Xencoded-by", "HERMES"),
                            ]
                        )
                    }
                    # Make a subprocess pipe to FFMPEG that streams in our frames and encode them into a video.
                    video_stream = ffmpeg.input(
                        "pipe:",  # type: ignore
                        format=input_stream_format,
                        pix_fmt=input_stream_pix_fmt,  # color format of piped input frames.
                        s="{}x{}".format(
                            frame_width, frame_height
                        ),  # size of frames from the sensor.
                        framerate=fps,
                        cpucount=self._spec.video_codec.num_cpu,
                        **self._spec.video_codec.input_options,
                    )
                    # TODO: use this to stream encoded video into a local file, and also as RTSP stream to the GUI.
                    # video_stream = ffmpeg.filter_multi_output
                    video_stream = ffmpeg.output(
                        video_stream,  # type: ignore
                        filename=filepath_video,
                        vcodec=self._spec.video_codec.codec_name,
                        pix_fmt=self._spec.video_codec.pix_format,
                        cpucount=self._spec.video_codec.num_cpu,  # prevent ffmpeg from suffocating the processor.
                        **self._spec.video_codec.output_options,
                        **metadata_dict,
                    )
                    video_stream = video_stream.global_args("-hide_banner")
                    video_writer: Popen = ffmpeg.run_async(video_stream, quiet=self._spec.is_quiet, pipe_stdin=True)  # type: ignore
                    # Store the writer.
                    self._video_writers.append(
                        (video_writer, streamer_name, device_name, stream_name)
                    )
                    num_writers += 1
        return num_writers

    def _init_files_audio(self) -> int:
        """[Not implemented] Create and initialize audio writers, one for each device.

        TODO: implement audio streaming info on the Stream object.

        Will fail if no FFmpeg installed.

        Raises:
            ValueError: When no supported codec specification was provided in config file.

        Returns:
            int: Number of initialized writers.
        """
        if self._spec.audio_codec is None:
            raise ValueError(
                "Must provide audio codec specification when streaming audio."
            )

        num_writers: int = 0
        for streamer_name, streamer in self._streams.items():
            for device_name, device_info in streamer.get_stream_info_all().items():
                for stream_name, stream_info in device_info.items():
                    # Skip non-audio streams.
                    if not stream_info["is_audio"]:
                        continue
                    # Create a unique file.
                    filename_base = "%s_%s" % (self._log_tag, device_name)
                    filename_audio = "%s.mp3" % (filename_base)
                    filepath_audio = os.path.join(self._spec.log_dir, filename_audio)
                    num_to_append = 0
                    while os.path.exists(filepath_audio):
                        num_to_append += 1
                        filename_audio = "%s_%02d.mp3" % (filename_base, num_to_append)
                        filepath_audio = os.path.join(
                            self._spec.log_dir, filename_audio
                        )

                    # Create an audio writer.
                    fps = stream_info["sampling_rate_hz"]
                    num_channels = stream_info["num_channels"]
                    input_stream_sample_fmt = stream_info["sample_format"]

                    metadata_dict = {
                        "metadata:g:%d" % i: "%s=%s" % (k, v)
                        for i, (k, v) in enumerate(
                            [
                                ("title", "/".join(self._spec.experiment.values())),
                                (
                                    "date",
                                    get_time_str(self._spec.log_time_s, "%Y-%m-%d"),
                                ),
                                (
                                    "comment",
                                    "HERMES multi-modal data acquisition system recording",
                                ),
                                *map(
                                    lambda tup: ("X%s" % tup[0], tup[1]),
                                    list(self._spec.experiment.items()),
                                ),
                                ("Xencoder", self._spec.audio_codec.codec_name),
                                ("Xencoded-by", "HERMES"),
                            ]
                        )
                    }
                    # Make a subprocess pipe to FFMPEG that streams in our frames and encode them into an audio.
                    audio_stream = ffmpeg.input(
                        "pipe:",  # type: ignore
                        ar=fps,
                        ac=num_channels,
                        cpucount=self._spec.audio_codec.num_cpu,
                        **self._spec.audio_codec.input_options,
                    )
                    # TODO: use this to stream encoded audio into a local file, and also as RTSP stream to the GUI.
                    # audio_stream = ffmpeg.filter_multi_output
                    audio_stream = ffmpeg.output(
                        audio_stream,  # type: ignore
                        filename=filepath_audio,
                        acodec=self._spec.audio_codec.codec_name,
                        sample_fmt=input_stream_sample_fmt,
                        cpucount=self._spec.audio_codec.num_cpu,  # prevent ffmpeg from suffocating the processor.
                        **self._spec.audio_codec.output_options,
                        **metadata_dict,
                    )
                    audio_stream = audio_stream.global_args("-hide_banner")
                    audio_writer: Popen = ffmpeg.run_async(audio_stream, quiet=self._spec.is_quiet, pipe_stdin=True)  # type: ignore
                    # Store the writer.
                    self._audio_writers.append(
                        (audio_writer, streamer_name, device_name, stream_name)
                    )
                    num_writers += 1
        return num_writers

    def _log_metadata_csv(self) -> None:
        """Add experiment metadata on the CSV files.

        TODO: validate logic.
        """
        for streamer_name, stream in self._streams.items():
            for device_name, device_info in stream.get_stream_info_all().items():
                # Get data notes for each stream.
                for stream_name, stream_info in device_info.items():
                    data_notes = stream_info["data_notes"]
                    if isinstance(data_notes, dict):
                        stream_metadata = data_notes
                    else:
                        stream_metadata = {"data_notes": data_notes}
                    stream_metadata = convert_dict_values_to_str(
                        stream_metadata, preserve_nested_dicts=False
                    )
                    # Write the stream-level metadata.
                    if self._csv_writer_metadata is not None:
                        self._csv_writer_metadata.write("\n")
                        self._csv_writer_metadata.write(
                            "Stream Name,%s" % (stream_name)
                        )
                        for meta_key, meta_value in stream_metadata.items():
                            self._csv_writer_metadata.write("\n")
                            self._csv_writer_metadata.write(
                                '%s,"%s"' % (str(meta_key), str(meta_value))
                            )
                        self._csv_writer_metadata.write("\n")

    def _log_metadata_hdf5(self) -> None:
        """Add experiment metadata on the HDF5 file."""
        file_metadata = convert_dict_values_to_str(
            {
                **self._spec.experiment,
                "Date": get_time_str(self._spec.log_time_s, "%Y-%m-%d"),
                "Time": get_time_str(self._spec.log_time_s, "%H-%M-%S"),
                "Comment": "HERMES multi-modal data acquisition system recording",
            },
            preserve_nested_dicts=False,
        )
        if self._hdf5_file is not None:
            file_group = self._hdf5_file["/"]
            file_group.attrs.update(file_metadata)
            # Add metadata per stream.
            # Flatten and prune the dictionary to make it HDF5 compatible.
            for streamer_name, stream in self._streams.items():
                # Add the class name.
                streamer_metadata = convert_dict_values_to_str(
                    {Stream.metadata_class_name_key: type(stream).__name__},
                    preserve_nested_dicts=False,
                )
                streamer_group = self._hdf5_file["/".join([streamer_name])]
                streamer_group.attrs.update(streamer_metadata)
                for device_name, device_info in stream.get_stream_info_all().items():
                    # NOTE: no per-device metadata for now.
                    # Get data notes for each stream.
                    for stream_name, stream_info in device_info.items():
                        try:
                            stream_group = self._hdf5_file[
                                "/".join([streamer_name, device_name, stream_name])
                            ]
                            data_notes = stream_info["data_notes"]
                            if isinstance(data_notes, dict):
                                stream_metadata = data_notes
                            else:
                                stream_metadata = {"Notes": data_notes}
                            stream_metadata = convert_dict_values_to_str(
                                stream_metadata, preserve_nested_dicts=False
                            )
                            stream_group.attrs.update(stream_metadata)
                        except KeyError:  # a writer was not created for this stream
                            pass

    def _log_metadata_video(self) -> None:
        """Add experiment metadata on the video files.

        Dummy method, metadata is written on instantiation.
        """
        pass

    def _log_metadata_audio(self) -> None:
        """Add experiment metadata on the audio files.

        Dummy method, metadata is written on instantiation.
        """
        pass

    def _log_metadata(self):
        """Convenience method to add metadata to all file types.

        Will include device-level metadata and any lower-level data notes.
        """
        self._log_metadata_csv()
        self._log_metadata_hdf5()
        self._log_metadata_video()
        self._log_metadata_audio()

    def _close_files_hdf5(self) -> None:
        """Flush/close the HDF5 file writer.

        Resizes datasets to remove extra empty rows.
        """
        if self._hdf5_file is not None:
            for streamer_name, stream in self._streams.items():
                for device_name, device_info in stream.get_stream_info_all().items():
                    for stream_name, stream_info in device_info.items():
                        try:
                            dataset: h5py.Dataset = self._hdf5_file["/".join([streamer_name, device_name, stream_name])]  # type: ignore
                        except KeyError:  # a dataset was not created for this stream
                            continue
                        starting_index = self._next_data_indices_hdf5[streamer_name][
                            device_name
                        ][stream_name]
                        ending_index = starting_index - 1
                        dataset.resize((ending_index + 1, *dataset.shape[1:]))
            self._hdf5_file.close()
            self._hdf5_file = None

    def _close_files_video(self) -> None:
        """Flush/close the video files writers."""
        for (video_writer, *_) in self._video_writers:
            video_writer.stdin.close()  # type: ignore
            if self._spec.is_quiet:
                video_writer.stderr.close()  # type: ignore
                video_writer.stdout.close()  # type: ignore
            video_writer.wait()
        self._video_writers = []

    def _close_files_csv(self) -> None:
        """Flush/close the CSV file writers."""
        for (stream_writer, *_) in self._csv_writers:
            stream_writer.close()
        self._csv_writers = []
        if self._csv_writer_metadata is not None:
            self._csv_writer_metadata.close()
            self._csv_writer_metadata = None

    def _close_files_audio(self) -> None:
        """Flush/close the audio file writers."""
        for (audio_writer, *_) in self._audio_writers:
            audio_writer.stdin.close()  # type: ignore
            if self._spec.is_quiet:
                audio_writer.stderr.close()  # type: ignore
                audio_writer.stdout.close()  # type: ignore
            audio_writer.wait()
        self._audio_writers = []

    def _close_files(self) -> None:
        """Convenience method to close all files writers."""
        self._close_files_csv()
        self._close_files_hdf5()
        self._close_files_video()
        self._close_files_audio()

    def _sync_write_hdf5(
        self, node_name: str, device_name: str, stream_name: str
    ) -> None:
        """Write provided data to the HDF5 file.

        Can be called during streaming (periodic writing) and post-experiment dumping.

        Args:
            node_name (str): Valid unique tag of the Node owning the data.
            device_name (str): Valid device tree name.
            stream_name (str): Valid sub-stream name.
        """
        if self._hdf5_file is not None:
            try:
                dataset: h5py.Dataset = self._hdf5_file["/".join([node_name, device_name, stream_name])]  # type: ignore
            except KeyError:  # a dataset was not created for this stream
                return
            new_data: Iterator[Any] = self._streams[node_name].pop_data(
                device_name=device_name,
                stream_name=stream_name,
                is_flush=self._is_flush,
            )
            for data in new_data:
                dataset_dtype: np.dtype = dataset.dtype
                # if dataset_dtype.char == 'S':
                #   data: str
                #   encoded_text = [data.encode("ascii", "ignore")]
                #   arr = np.array(encoded_text, ndmin=1)
                # else:
                arr = np.array(data, ndmin=1)
                num_elements = 1 if arr.shape == dataset.shape[1:] else arr.shape[0]
                # Extend the dataset as needed while iterating over the 'new_data'.
                start_index = self._next_data_indices_hdf5[node_name][device_name][
                    stream_name
                ]
                # Expand the dataset if needed.
                if not (start_index + num_elements < len(dataset)):
                    dataset.resize(
                        (
                            len(dataset) + max(self._hdf5_log_length_increment, num_elements),
                            *dataset.shape[1:],
                        )
                    )
                dataset[start_index : start_index + num_elements, :] = arr
                # Write the new entries.
                # Update the next starting index to use.
                start_index += num_elements
                self._next_data_indices_hdf5[node_name][device_name][
                    stream_name
                ] = start_index
            # Flush the file with the new data.
            self._hdf5_file.flush()

    def _sync_write_video(
        self, video_writer: Popen, node_name: str, device_name: str, stream_name: str
    ):
        """Write provided data to the video files.

        Can be called during streaming (periodic writing) and post-experiment dumping.

        Args:
            video_writer (Popen): FFmpeg writer corresponding to the video.
            node_name (str): Valid unique tag of the Node owning the data.
            device_name (str): Valid device tree name.
            stream_name (str): Valid sub-stream name.
        """
        new_data: Iterator[tuple[bytes, bool, int]] = self._streams[node_name].pop_data(
            device_name=device_name, stream_name=stream_name, is_flush=self._is_flush
        )
        for frame_buffer, is_keyframe, frame_index in new_data:
            video_writer.stdin.write(frame_buffer)  # type: ignore

    def _sync_write_csv(
        self,
        csv_writer: TextIOWrapper,
        node_name: str,
        device_name: str,
        stream_name: str,
    ) -> None:
        """Write provided data to the CSV file.

        Can be called during streaming (periodic writing) and post-experiment dumping.

        Args:
            csv_writer (TextIOWrapper): Text file writer corresponding to the stream.
            node_name (str): Valid unique tag of the Node owning the data.
            device_name (str): Valid device tree name.
            stream_name (str): Valid sub-stream name.
        """
        new_data: Iterator[Any] = self._streams[node_name].pop_data(
            device_name=device_name, stream_name=stream_name, is_flush=self._is_flush
        )
        for data_to_write in new_data:
            # Create a list of column entries to write.
            # Note that they should match the heading order in _init_writing_csv().
            if isinstance(data_to_write, np.ndarray):
                to_write = list(np.atleast_1d(data_to_write.reshape(1, -1).squeeze()))
            elif isinstance(data_to_write, (list, tuple)):
                to_write = data_to_write
            else:
                to_write = list(data_to_write)
            # Write the new row.
            csv_writer.write("\n")
            csv_writer.write(",".join([str(x) for x in to_write]))
        csv_writer.flush()

    def _sync_write_audio(
        self, audio_writer: Popen, node_name: str, device_name: str, stream_name: str
    ):
        """Write provided data to the audio files.

        Can be called during streaming (periodic writing) and post-experiment dumping.

        Args:
            audio_writer (Popen): FFmpeg writer corresponding to the audio.
            node_name (str): Valid unique tag of the Node owning the data.
            device_name (str): Valid device tree name.
            stream_name (str): Valid sub-stream name.
        """
        new_data: Iterator[Any] = self._streams[node_name].pop_data(
            device_name=device_name, stream_name=stream_name, is_flush=self._is_flush
        )
        for frame in new_data:
            audio_writer.stdin.write(frame)  # type: ignore

    def _write_hdf5(self) -> None:
        """Convenience wrapper to write new data for multiple HDF5 streams.

        Writes all new text data to a single HDF5 file.
        """
        for streamer_name, stream in self._streams.items():
            for device_name, device_info in stream.get_stream_info_all().items():
                for stream_name, stream_info in device_info.items():
                    self._sync_write_hdf5(
                        node_name=streamer_name,
                        device_name=device_name,
                        stream_name=stream_name,
                    )

    async def _write_video(
        self, video_writer: Popen, node_name: str, device_name: str, stream_name: str
    ):
        """Coroutine asynchronous wrapper for multiple video streams.

        Wraps synchronous IO of each video file into an asynchronous pool of coroutines.

        Args:
            video_writer (Popen): FFmpeg writer corresponding to the video.
            node_name (str): Valid unique tag of the Node owning the data.
            device_name (str): Valid device tree name.
            stream_name (str): Valid sub-stream name.
        """
        await asyncio.get_event_loop().run_in_executor(
            self._thread_pool,
            lambda: self._sync_write_video(
                video_writer=video_writer,
                node_name=node_name,
                device_name=device_name,
                stream_name=stream_name,
            ),
        )

    async def _write_csv(
        self,
        csv_writer: TextIOWrapper,
        node_name: str,
        device_name: str,
        stream_name: str,
    ):
        """Coroutine asynchronous wrapper for multiple CSV streams.

        Wraps synchronous IO of each CSV file into an asynchronous pool of coroutines.

        Args:
            csv_writer (TextIOWrapper): Text file writer corresponding to the stream.
            node_name (str): Valid unique tag of the Node owning the data.
            device_name (str): Valid device tree name.
            stream_name (str): Valid sub-stream name.
        """
        await asyncio.get_event_loop().run_in_executor(
            self._thread_pool,
            lambda: self._sync_write_csv(
                csv_writer=csv_writer,
                node_name=node_name,
                device_name=device_name,
                stream_name=stream_name,
            ),
        )

    async def _write_audio(
        self, audio_writer: Popen, node_name: str, device_name: str, stream_name: str
    ):
        """Coroutine asynchronous wrapper for multiple audio streams.

        Wraps synchronous IO of each video file into an asynchronous pool of coroutines.

        Args:
            audio_writer (Popen): FFmpeg writer corresponding to the audio.
            node_name (str): Valid unique tag of the Node owning the data.
            device_name (str): Valid device tree name.
            stream_name (str): Valid sub-stream name.
        """
        await asyncio.get_event_loop().run_in_executor(
            self._thread_pool,
            lambda: self._sync_write_audio(
                audio_writer=audio_writer,
                node_name=node_name,
                device_name=device_name,
                stream_name=stream_name,
            ),
        )

    async def _write_files_hdf5(self):
        """Convenience method to asynchronously write all HDF5 data to file.

        Will launch coroutine concurrently with other IO writer types.
        """
        await asyncio.get_event_loop().run_in_executor(
            self._thread_pool, lambda: self._write_hdf5()
        )

    async def _write_files_video(self):
        """Convenience method to asynchronously write all video data to files.

        Will launch coroutines concurrently with other IO writer types.
        """
        tasks = []
        for (
            video_writer,
            streamer_name,
            device_name,
            stream_name,
        ) in self._video_writers:
            tasks.append(
                self._write_video(
                    video_writer=video_writer,
                    node_name=streamer_name,
                    device_name=device_name,
                    stream_name=stream_name,
                )
            )
        await asyncio.gather(*tasks)

    async def _write_files_csv(self):
        """Convenience method to asynchronously write all CSV data to files.

        Will launch coroutine concurrently with other IO writer types.
        """
        tasks = []
        for stream_writer, streamer_name, device_name, stream_name in self._csv_writers:
            tasks.append(
                self._write_csv(
                    csv_writer=stream_writer,
                    node_name=streamer_name,
                    device_name=device_name,
                    stream_name=stream_name,
                )
            )
        await asyncio.gather(*tasks)

    async def _write_files_audio(self):
        """Convenience method to asynchronously write all audio data to files.

        Will launch coroutine concurrently with other IO writer types.
        """
        tasks = []
        for (
            audio_writer,
            streamer_name,
            device_name,
            stream_name,
        ) in self._audio_writers:
            tasks.append(
                self._write_audio(
                    audio_writer=audio_writer,
                    node_name=streamer_name,
                    device_name=device_name,
                    stream_name=stream_name,
                )
            )
        await asyncio.gather(*tasks)

    ##########################
    ###### DATA LOGGING ######
    ##########################
    async def _log_data(self) -> None:
        """Trigger release of AsyncIO resources used for writing files.

        Polls data from each Node periodically or all at once.

        The poll period is set by self._stream_period_s.

        Will loop until self._is_streaming is False, and then
        will do one final fetch/log if self._is_flush is True.

        Assert self._is_streaming and deassert self._is_flush for streaming.
            To finish, deassert self._is_streaming to False and assert self._is_flush.

        Deassert self._is_streaming and assert self._is_flush to dump record.
            The thread will be inactive until terminated.
            User is responsible to provision sufficient memory.
        """
        last_log_time_s = None
        # Set at the beginning of the iteration if _is_flush is externally modified to indicate cleanup and exit,
        #   to catch case where external command to flush happened while some of streamers already saved part of available data.
        is_flush_all_in_current_iteration = False
        while (self._is_streaming or self._is_flush) and not self._is_finished:
            # Wait until it is time to write new data, which is either:
            #  1. This is the first iteration.
            #  2. It has been at least self._stream_period_s since the last write.
            #  3. Periodic logging has been deactivated.
            while (
                last_log_time_s is not None
                and (
                    time_to_next_period := (
                        last_log_time_s + self._spec.stream_period_s - get_time()
                    )
                )
                > 0
                and self._is_streaming
            ):
                # Will wake up periodically to check if the experiment had been ended.
                #   Will proceed only if time for next logging or if experiment ended.
                await asyncio.sleep(min(1, time_to_next_period))
            # If running Logger in dump mode, wait until _is_flush is set externally.
            if not self._is_streaming and not self._is_flush:
                continue
            # Update the last log time now, before the write actually starts.
            # This will keep the log period more consistent; otherwise, the amount
            #   of time it takes to perform the write would be added to the log period.
            #   This would compound over time, leading to longer delays and more data to write each time.
            #   This becomes more severe as the write duration increases (e.g. videos).
            last_log_time_s = get_time()
            # If the log should be flushed, record that it is happening during this iteration for ALL streamers.
            if self._is_flush:
                is_flush_all_in_current_iteration = True
            # Delegate file writing to each AsyncIO method that manages corresponding stream type writing.
            tasks = []
            if self._spec.stream_hdf5:
                tasks.append(self._write_files_hdf5())
            if self._spec.stream_video:
                tasks.append(self._write_files_video())
            if self._spec.stream_csv:
                tasks.append(self._write_files_csv())
            if self._spec.stream_audio:
                tasks.append(self._write_files_audio())
            # Execute all file writing concurrently.
            await asyncio.gather(*tasks)
            # If stream-logging is disabled, but a final flush had been requested,
            #   record that the flush is complete so streaming can really stop now.
            # Note that it also checks whether the flush was configured to happen for all streamers during this iteration.
            #   Especially if a lot of data was being written (such as with video data),
            #   the self._is_flush flag may have been set sometime during the data writing.
            #   In that case, all streamers would not have known to flush data and some data may be omitted.
            # flushing_log set True when _is_flush was set before any streamer saved its data chunk, to make ure nothing is left behind.
            if (
                (not self._is_streaming)
                and self._is_flush
                and is_flush_all_in_current_iteration
            ):
                self._is_finished = True
        # Log metadata.
        self._log_metadata()
        # Save and close the files.
        self._close_files()



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

from abc import ABC, abstractmethod

import copy
from collections import OrderedDict, deque
from typing import Any, Dict, Iterable, Iterator, Mapping
from threading import Lock

from hermes.utils.time_utils import get_time
from hermes.utils.types import (
    VideoFormatEnum,
    DataFifoDict,
    DeviceLockDict,
    ExtraDataInfoDict,
    NewDataDict,
    StreamInfoDict,
)


class Stream(ABC):
    """An abstract class to store data of a Node.

    Tree-like structure of FIFO buffers. May contain multiple sub-streams
    for a single device, e.g. acceleration and gyroscope of an IMU.

    Data for sub-streams under the same device tree arrives as a single packet.
    Packets containing decoupled data are better treated as independent device trees.

    Uses a lock for each device tree to delegate access to the start of the FIFO:
    ensures high performance from parallel acquisition, processing, and logging blocks.
    This allows the end of the FIFO to be saved and discarded by the Storage.

    Will store the class name of each sensor in HDF5 metadata.

    Can periodically clear old data (if needed).
    """

    metadata_class_name_key = "Stream class name"
    metadata_data_headings_key = "Data headings"

    _data: DataFifoDict
    _streams_info: StreamInfoDict
    _locks: DeviceLockDict

    def __init__(self) -> None:
        self._data = dict()
        self._streams_info = dict()
        self._locks = dict()

    ############################
    ###### INTERFACE FLOW ######
    ############################
    @abstractmethod
    def get_fps(self) -> dict[str, float | None]:
        """Get effective frame rate of this unique stream's captured data.

        Subject to expected transmission delay and throughput limitation.
        Computed based on how fast data becomes available to the data structure.
        Used to measure the performance of the system - local or remote nodes.

        Returns:
            dict[str, float | None]: Mapping of measured FPS to stream names.
        """
        pass

    #############################
    ###### GETTERS/SETTERS ######
    #############################
    def add_stream(
        self,
        device_name: str,
        stream_name: str,
        data_type: str,
        sample_size: Iterable[int],
        sampling_rate_hz: float = 0.0,
        is_measure_rate_hz: bool = False,
        data_notes: Mapping[str, str] = {},
        is_video: bool = False,
        color_format: VideoFormatEnum | None = None,
        is_audio: bool = False,
        timesteps_before_solidified: int = 0,
        extra_data_info: ExtraDataInfoDict = {},
    ) -> None:
        """Add a new sub-stream to an existing device tree or creates new.

        Will by default add a stream for each device to mark each captured sample
        with the host's time-of-arrival.

        Args:
            device_name (str): Device tree name. Will autocreate if doesn't exist.
            stream_name (str): Unique sub-stream name under this device tree.
            data_type (str): Fixed data type expected in the sub-stream.
            sample_size (Iterable[int]): An interable of dimensions of given data type in each captured sample.
            sampling_rate_hz (float, optional): Expected sampling frequency of the signal. Defaults to `0.0`.
            is_measure_rate_hz (bool, optional): Whether to compute the effective sampling frequency. Defaults to `False`.
            data_notes (Mapping[str, str], optional): Mapping of streams to notes for Storage to use in file metadata. Defaults to `{}`.
            is_video (bool, optional): Whether it is a video stream. Defaults to `False`.
            color_format (VideoFormatEnum | None, optional): One of the supported identifiers (see `types.py`). Defaults to `None`.
            is_audio (bool, optional): Whether it is an audio stream. Defaults to `False`.
            timesteps_before_solidified (int, optional): How many most recent samples to keep in memory before flushing. Defaults to `0`.
            extra_data_info (ExtraDataInfoDict, optional): Additional mapping that will be streamed along with data,
                with at least 'data_type' and 'sample_size'. Defaults to `{}`.

        Raises:
            ValueError: If stream name is not unique or is reserved.
        """
        if stream_name == "process_time_s":
            raise ValueError("'process_time_s' is reserved for Stream internal use.")
        self._add_stream(
            device_name=device_name,
            stream_name=stream_name,
            data_type=data_type,
            sample_size=sample_size,
            sampling_rate_hz=sampling_rate_hz,
            is_measure_rate_hz=is_measure_rate_hz,
            data_notes=data_notes,
            is_video=is_video,
            color_format=color_format,
            is_audio=is_audio,
            timesteps_before_solidified=timesteps_before_solidified,
            extra_data_info=extra_data_info,
        )
        if "process_time_s" not in self._data[device_name]:
            self._add_stream(
                device_name=device_name,
                stream_name="process_time_s",
                data_type="float64",
                sample_size=(1,),
                data_notes=OrderedDict(
                    [
                        (
                            "Description",
                            "Time of arrival of the data point to the host PC, "
                            "to be used for aligned idexing of data between distributed hosts.",
                        )
                    ]
                ),
            )

    def _add_stream(
        self,
        device_name: str,
        stream_name: str,
        data_type: str,
        sample_size: Iterable[int],
        sampling_rate_hz: float = 0.0,
        is_measure_rate_hz: bool = False,
        data_notes: Mapping[str, str] = {},
        is_video: bool = False,
        color_format: VideoFormatEnum | None = None,
        is_audio: bool = False,
        timesteps_before_solidified: int = 0,
        extra_data_info: ExtraDataInfoDict = {},
    ) -> None:
        """[Internal] Underlying logic for adding a stream.

        Raises:
            KeyError: If supplied color format is not supported or misspelled.
        """
        self._locks.setdefault(device_name, Lock())
        self._streams_info.setdefault(device_name, dict())
        if not isinstance(sample_size, Iterable):
            sample_size = [sample_size]
        self._streams_info[device_name][stream_name] = dict(
            [
                ("data_type", data_type),
                ("sample_size", sample_size),
                ("data_notes", data_notes),
                ("sampling_rate_hz", "%.2f" % sampling_rate_hz),
                ("is_measure_rate_hz", is_measure_rate_hz),
                ("is_video", is_video),
                ("is_audio", is_audio),
                ("timesteps_before_solidified", timesteps_before_solidified),
                ("extra_data_info", extra_data_info),
            ]
        )
        # Record color formats to use by FFmpeg, for saving and displaying frames.
        if is_video:
            try:
                if color_format is not None:
                    self._streams_info[device_name][stream_name][
                        "format"
                    ] = color_format.value.format
                    self._streams_info[device_name][stream_name][
                        "color"
                    ] = color_format.value.color
                else:
                    raise KeyError
            except KeyError:
                print(
                    "Color format %s is not supported when specifying video frame pixel color format on Stream."
                    % color_format
                )
        # Some metadata to keep track of during running to measure the actual frame rate.
        if is_measure_rate_hz:
            # Set at start actual rate equal to desired sample rate
            self._streams_info[device_name][stream_name][
                "actual_rate_hz"
            ] = sampling_rate_hz
            # Create a circular buffer of 1 second, w.r.t. desired sample rate
            circular_buffer_len: int = max(round(sampling_rate_hz), 1)
            self._streams_info[device_name][stream_name]["dt_circular_buffer"] = list(
                [1 / sampling_rate_hz] * circular_buffer_len
            )
            self._streams_info[device_name][stream_name]["dt_circular_index"] = 0
            self._streams_info[device_name][stream_name]["dt_running_sum"] = 1.0
            self._streams_info[device_name][stream_name]["old_toa"] = get_time()
        self.clear_data(device_name, stream_name)

    def append_data(self, process_time_s: float, data: NewDataDict) -> None:
        """Thread-safe append of new sample to the stream.

        Locks the device tree of the sub-stream, to avoid immutability error
        in reverse iterator of the GUI thread when trying to peek N newest
        samples of the stream while new are written.

        Args:
            process_time_s (float): Time-of-processing of the sample.
            data (NewDataDict): Newly processed sample.
        """
        for device_name, streams_data in data.items():
            if streams_data is not None:
                self._locks[device_name].acquire()
                for stream_name, stream_data in streams_data.items():
                    self._append(device_name, stream_name, stream_data)
                self._append(device_name, "process_time_s", process_time_s)
                self._locks[device_name].release()

    def _append(self, device_name: str, stream_name: str, data: Any) -> None:
        """[Internal] Non thread-safe append of new sample."""
        self._data[device_name][stream_name].append(data)

        # If stream set to measure actual fps
        # TODO: cleanup to use a fixed-length Deque instead.
        if self._streams_info[device_name][stream_name]["is_measure_rate_hz"]:
            # Make intermediate variables for current and previous samples' time-of-arrival
            new_toa = get_time()
            old_toa = self._streams_info[device_name][stream_name]["old_toa"]
            # Record the new arrival time for the next iteration
            self._streams_info[device_name][stream_name]["old_toa"] = new_toa
            # Update the running sum of time increments of the circular buffer
            oldest_dt = self._streams_info[device_name][stream_name][
                "dt_circular_buffer"
            ][self._streams_info[device_name][stream_name]["dt_circular_index"]]
            newest_dt = new_toa - old_toa
            self._streams_info[device_name][stream_name]["dt_running_sum"] += (
                newest_dt - oldest_dt
            )
            # Put current time increment in place of the oldest one in the circular buffer
            self._streams_info[device_name][stream_name]["dt_circular_buffer"][
                self._streams_info[device_name][stream_name]["dt_circular_index"]
            ] = newest_dt
            # Move the index in the circular fashion
            self._streams_info[device_name][stream_name]["dt_circular_index"] = (
                self._streams_info[device_name][stream_name]["dt_circular_index"] + 1
            ) % len(self._streams_info[device_name][stream_name]["dt_circular_buffer"])
            # Refresh the actual frame rate information
            self._streams_info[device_name][stream_name]["actual_rate_hz"] = (
                len(self._streams_info[device_name][stream_name]["dt_circular_buffer"])
                / self._streams_info[device_name][stream_name]["dt_running_sum"]
            )

    def pop_data(
        self,
        device_name: str,
        stream_name: str,
        num_oldest_to_pop: int | None = None,
        is_flush: bool = False,
    ) -> Iterator[Any]:
        """Wrap all samples ready to be popped in an iterator oldest->newest.

        Used by Storage to flush data to disk.
        Popped data is cleared from memory.
        Thread-safe without locks while appending new data.

        Args:
            device_name (str): Valid device tree name.
            stream_name (str): Valid sub-stream name.
            num_oldest_to_pop (int | None, optional): Number of samples to pop. Defaults to `None`.
            is_flush (bool, optional): Whether to pop all data in the stream, regardless of timesteps_before_solidified. Defaults to `False`.

        Yields:
            Iterator[Any]: Iterator over poppable oldest->newest samples.
        """
        # O(1) complexity to check length of a Deque.
        num_available: int = len(self._data[device_name][stream_name])
        # Can pop all available data, except what must be kept peekable.
        num_poppable: int = (
            num_available
            - self._streams_info[device_name][stream_name][
                "timesteps_before_solidified"
            ]
        )
        # If experiment ended, flush all available data from the Stream.
        if is_flush:
            num_oldest_to_pop = num_available
        elif num_oldest_to_pop is None:
            num_oldest_to_pop = num_poppable
        else:
            num_oldest_to_pop = min(num_oldest_to_pop, num_poppable)
        # Iterate through the doubly-linked list, clearing popped data, while new data is added to it.
        num_popped: int = 0
        while num_popped < num_oldest_to_pop:
            yield self._data[device_name][stream_name].popleft()
            num_popped += 1

    def peek_data_new(
        self, device_name: str, stream_name: str, num_newest_to_peek: int | None = None
    ) -> Iterator[Any]:
        """Wrap N newest samples in an iterator to peek.

        Will lock the device tree of the sub-stream to prevent appends muttating the iterator.
        Will allow popping of the oldest data (e.g. for Storage to flush).
        Peeking and popping ranges are protected by `timesteps_before_solidified`.

        Args:
            device_name (str): Valid device tree name.
            stream_name (str): Valid sub-stream name.
            num_newest_to_peek (int | None, optional): Number of samples to peek, if less than `timesteps_before_solidified`. Defaults to `None`.

        Yields:
            Iterator[Any]: Iterator over peekable newest samples.
        """
        self._locks[device_name].acquire()
        num_peekable: int = min(
            self._streams_info[device_name][stream_name]["timesteps_before_solidified"],
            len(self._data[device_name][stream_name]),
        )
        if num_newest_to_peek is None:
            num_newest_to_peek = num_peekable
        else:
            num_newest_to_peek = min(
                num_newest_to_peek,
                self._streams_info[device_name][stream_name][
                    "timesteps_before_solidified"
                ],
            )
        num_peeked: int = 0
        # Get an iterator to traverse the linked list from the write end (newest data) -> O(1).
        stream_reversed = reversed(self._data[device_name][stream_name])
        while num_peeked < num_newest_to_peek:
            yield next(stream_reversed)
            num_peeked += 1
        self._locks[device_name].release()

    def clear_data(
        self, device_name: str, stream_name: str, num_oldest_to_clear: int | None = None
    ) -> None:
        """Clear data in a stream.

        Initializes the sub-stream if it doesn't exist.
        Optionally can clear N oldest samples.

        Args:
            device_name (str): Valid device tree name.
            stream_name (str): Valid sub-stream name.
            num_oldest_to_clear (int | None, optional): Number of oldest samples to clear. Defaults to `None`.
        """
        # Create the device/stream entry if it doesn't exist, else clear it.
        self._data.setdefault(device_name, OrderedDict())
        if stream_name not in self._data[device_name]:
            self._data[device_name][stream_name] = deque()
        elif num_oldest_to_clear is not None:
            # Clearing up to a point in the Deque.
            # Wait until neither Node, nor GUI, append or peek newest data, respectively,
            #   only if clearing past their operating area.
            num_clearable: int = (
                len(self._data[device_name][stream_name])
                - self._streams_info[device_name][stream_name][
                    "timesteps_before_solidified"
                ]
            )
            is_to_lock: bool = not (num_oldest_to_clear < num_clearable)
            num_cleared: int = 0
            if is_to_lock:
                self._locks[device_name].acquire()
            while num_cleared < num_oldest_to_clear:
                self._data[device_name][stream_name].popleft()
                num_cleared += 1
            if is_to_lock:
                self._locks[device_name].release()
        else:
            # Clearing the whole Deque.
            # Wait until neither Node, nor GUI, append or peek newest data, respectively.
            self._locks[device_name].acquire()
            self._data[device_name][stream_name].clear()
            self._locks[device_name].release()

    def clear_data_all(self) -> None:
        """Clear all sub-streams from all device trees."""
        for device_name, device_info in self._streams_info.items():
            for stream_name, stream_info in device_info.items():
                self.clear_data(device_name, stream_name)

    def get_num_devices(self) -> int:
        """Get the number of asynchronous device trees.

        Returns:
            int: Number of device trees.
        """
        return len(self._streams_info)

    def get_device_names(self) -> list[str]:
        """Get the names of the asynchronous device trees.

        Returns:
            list[str]: Names of device trees.
        """
        return list(self._streams_info.keys())

    def get_stream_names(self, device_name: str | None = None) -> list[str]:
        """Get the names of sub-streams in a device tree.

        If device_name is None, will assume streams are the same for every device.

        Args:
            device_name (str | None, optional): Name of the device tree to query. Defaults to `None`.

        Returns:
            list[str]: Names of sub-streams in a device tree.
        """
        if device_name is None:
            device_name = self.get_device_names()[0]
        return list(self._streams_info[device_name].keys())

    def get_stream_info(self, device_name: str, stream_name: str) -> Dict[str, Any]:
        """Get metadata of a sub-stream.

        Args:
            device_name (str): Valid device tree name.
            stream_name (str): Valid sub-stream name.

        Returns:
            Dict[str, Any]: Metadata dictionary with keys:
                data_type
                is_video
                is_audio
                sample_size
                sampling_rate_hz
                timesteps_before_solidified
                extra_data_info
                data_notes
                if is_measure_rate_hz:
                actual_rate_hz
                dt_circular_buffer
                dt_circular_index
                dt_running_sum
                old_toa
        """
        return self._streams_info[device_name][stream_name]

    def get_stream_info_all(self) -> StreamInfoDict:
        """Get metadata of all sub-streams.

        Returns:
            StreamInfoDict: Nested dictionary of metadata, with device trees and sub-streams as keys.
        """
        return copy.deepcopy(self._streams_info)

    def _get_fps(self, device_name: str, stream_name: str) -> float | None:
        """[Internal] Retrieve the effective sampling rate of a signal, if recorded.

        Records and refreshes rolling statistics on each data structure append over 1-second windows.

        Args:
            device_name (str): Valid device tree name.
            stream_name (str): Valid sub-stream name.

        Returns:
            float | None: Measured acquisition sampling rate of the sub-stream.
        """
        if self._streams_info[device_name][stream_name]["is_measure_rate_hz"]:
            return self._streams_info[device_name][stream_name]["actual_rate_hz"]
        else:
            return None



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

import argparse


def validate_ip4(s: str) -> str:
    """Validates whether parsed value is a valid IPv4 address.

    Args:
        s (str): Parsed text value to test for validity as IPv4 address.

    Returns:
        str: Unchanged text value to use as IP address.

    Raises:
        argparse.ArgumentTypeError: If the parsed value is not a valid IPv4 address.
    """
    try:
        a = s.split(".")
        assert len(a) == 4
        for x in a:
            assert x.isdigit()
            n = int(x)
            assert n >= 0 and n <= 255
        return s
    except:
        raise argparse.ArgumentTypeError("Not a valid IPv4 address: ", s)


def validate_path(s: str) -> str:
    """Validates whether parsed file path exists.

    Args:
        s (str): File path to parse.

    Returns:
        str: Unchanged path if it exists.

    Raises:
        argparse.ArgumentTypeError: If the provided path does not exist.
    """
    try:
        return s
    except:
        raise argparse.ArgumentTypeError("Invalid path to config file: ", s)


def parse_type(s: str) -> int | float | bool | str:
    """Parsing utility to convert data types to native Python objects.
    
    Args:
        s (str): Text to convert to the corresponding data type.

    Returns:
        int | float | bool | str: Parsed primitive data type. 
    """
    if s.isdigit():
        return int(s)
    elif s == "True":
        return True
    elif s == "False":
        return False
    else:
        try:
            return float(s)
        except:
            return s


class ParseExperimentKwargs(argparse.Action):
    """Parsing object for experiment details specification."""
    def __call__(self, parser, namespace, values, option_string=None):
        if isinstance(values, (list, tuple)):
            setattr(namespace, self.dest, dict())
            for value in values:
                key, value = value.split("=")
                getattr(namespace, self.dest)[key] = value


class ParseStorageKwargs(argparse.Action):
    """Parsing object for `Storage` specification."""
    def __call__(self, parser, namespace, values, option_string=None):
        if isinstance(values, (list, tuple)):
            setattr(namespace, self.dest, dict())
            for value in values:
                if "=" in value:
                    key, val = value.split("=")
                    getattr(namespace, self.dest)[key] = parse_type(val)
                else:
                    getattr(namespace, self.dest)[value] = True


class ParseNodeKwargs(argparse.Action):
    """Parsing object for `Node` specifications."""
    def __call__(self, parser, namespace, values, option_string=None):
        if isinstance(values, (list, tuple)):
            new_items = list()
            # Parse the input values as a dictionary
            id = -1
            for value in values:
                key, val = value.split("=")
                if key == "class":
                    id += 1
                    new_items.append(dict([(key, val)]))
                elif ";" in val:
                    new_items[id][key] = dict()
                    for pair_str in val.split(";"):
                        k, v = pair_str.split(":")
                        new_items[id][key][k] = v
                elif "," in val:
                    new_items[id][key] = list(map(parse_type, val.split(",")))
                else:
                    new_items[id][key] = parse_type(val)
            # Extend the list with the new dictionary
            items = getattr(namespace, self.dest, list())
            items.extend(new_items)
            setattr(namespace, self.dest, items)



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

from abc import abstractmethod
from collections import OrderedDict
import threading
import zmq

from hermes.utils.time_utils import get_time
from hermes.utils.di_utils import search_module_class
from hermes.utils.msgpack_utils import deserialize, serialize
from hermes.utils.zmq_utils import (
    CMD_END,
    CMD_EXIT,
    DNS_LOCALHOST,
    PORT_BACKEND,
    PORT_FRONTEND,
    PORT_SYNC_HOST,
    PORT_KILL,
)
from hermes.utils.types import LoggingSpec

from hermes.base.stream import Stream
from hermes.base.storage.storage import Storage
from hermes.base.nodes.node import Node
from hermes.base.nodes.pipeline_interface import PipelineInterface
from hermes.base.nodes.producer_interface import ProducerInterface


class Pipeline(PipelineInterface, Node):
    """An abstract class to interface with a data-producing worker."""

    def __init__(
        self,
        host_ip: str,
        stream_out_spec: dict,
        stream_in_specs: list[dict],
        logging_spec: LoggingSpec,
        is_async_generate: bool = False,
        port_pub: str = PORT_BACKEND,
        port_sub: str = PORT_FRONTEND,
        port_sync: str = PORT_SYNC_HOST,
        port_killsig: str = PORT_KILL,
    ) -> None:
        """Constructor of the Pipeline parent class.

        Args:
            host_ip (str): IP address of the local master Broker.
            stream_out_spec (dict): Mapping of corresponding Stream object parameters to user-defined configuration values.
            stream_in_specs (list[dict]): List of mappings of user-configured incoming modalities.
            logging_spec (LoggingSpec): Specification of what and how to store.
            is_async_generate (bool, optional): Whether the Pipeline produces data asynchronously, in parallel to what is fed into it. Defaults to `False`.
            port_pub (str, optional): Local port to publish to for local master Broker to relay. Defaults to `PORT_BACKEND`.
            port_sub (str, optional): Local port to subscribe to for incoming relayed data from the local master Broker. Defaults to `PORT_FRONTEND`.
            port_sync (str, optional): Local port to listen to for local master Broker's startup coordination. Defaults to `PORT_SYNC_HOST`.
            port_killsig (str, optional): Local port to listen to for local master Broker's termination signal. Defaults to `PORT_KILL`.
        """
        super().__init__(
            host_ip=host_ip,
            port_sync=port_sync,
            port_killsig=port_killsig,
            ref_time=logging_spec.ref_time_s,
        )
        self._port_pub = port_pub
        self._port_sub = port_sub
        self._is_async_generate = is_async_generate
        self._is_more_data_in = True
        self._is_more_data_out = True
        self._publish_fn = lambda tag, **kwargs: None

        # Data structure for keeping track of the Pipeline's output data.
        self._out_stream: Stream = self.create_stream(stream_out_spec)

        # Instantiate all desired Streams that the Pipeline will process.
        self._in_streams: OrderedDict[str, Stream] = OrderedDict()
        self._poll_data_fn = self._poll_data_packets
        self._on_poll_fn = self._on_poll_in_out if self._is_async_generate else self._on_poll_in_only
        self._is_producer_ended: OrderedDict[str, bool] = OrderedDict()

        for stream_spec in stream_in_specs:
            module_name: str = stream_spec["package"]
            class_name: str = stream_spec["class"]
            specs: dict = stream_spec["settings"]
            # Create the stream datastructure.
            class_type: type[ProducerInterface] | type[PipelineInterface] = search_module_class(module_name, class_name)  # type: ignore
            class_object: Stream = class_type.create_stream(specs)
            self._in_streams.setdefault(class_type._log_source_tag(), class_object)
            self._is_producer_ended.setdefault(class_type._log_source_tag(), False)

        # Create the data storing object.
        self._storage = Storage(self._log_source_tag(), logging_spec)

        # Launch datalogging thread with reference to the Stream objects, to save Pipeline's outputs and inputs.
        self._storage_thread = threading.Thread(
            target=self._storage,
            args=(
                OrderedDict(
                    [
                        (self._log_source_tag(), self._out_stream),
                        *list(self._in_streams.items()),
                    ]
                ),
            ),
        )
        self._storage_thread.start()

    def _publish(self, tag: str, **kwargs) -> None:
        """Common method to save and publish the captured sample.

        Best to deal with data structure (threading primitives) AFTER handing off packet to ZeroMQ.
        That way network thread can already start processing the packet.

        Args:
            tag (str): Uniquely identifying key for the modality to label data for message exchange.
        """
        self._publish_fn(tag, **kwargs)

    def _initialize(self):
        super()._initialize()

        # Socket to publish processed data and log.
        self._pub: zmq.SyncSocket = self._ctx.socket(zmq.PUB)
        self._pub.connect("tcp://%s:%s" % (DNS_LOCALHOST, self._port_pub))

        # Socket to subscribe to other Producers.
        self._sub: zmq.SyncSocket = self._ctx.socket(zmq.SUB)
        self._sub.connect("tcp://%s:%s" % (DNS_LOCALHOST, self._port_sub))

        # Subscribe to topics for each mentioned local and remote streamer
        for tag in self._in_streams.keys():
            self._sub.subscribe(tag)

    def _activate_data_poller(self) -> None:
        self._poller.register(self._sub, zmq.POLLIN)
        if self._is_async_generate:
            self._poller.register(self._pub, zmq.POLLOUT)

    def _on_poll_in_only(self, poll_res: tuple[list[zmq.SyncSocket], list[int]]) -> None:
        """Callback to handle incoming data only.

        Args:
            poll_res: Result of zmq.Poller.poll() call.
        """
        if self._sub in poll_res[0]:
            self._poll_data_fn()

    def _on_poll_in_out(self, poll_res: tuple[list[zmq.SyncSocket], list[int]]) -> None:
        """Callback to handle incoming data and asynchronously generated internal outgoing data.

        Args:
            poll_res: Result of zmq.Poller.poll() call.
        """
        if self._sub in poll_res[0]:
            self._poll_data_fn()
        if self._pub in poll_res[0]:
            self._generate_data()

    def _on_poll(self, poll_res):
        # Receiving a modality packet, process until all data sources sent 'END' packet.
        self._on_poll_fn(poll_res)
        super()._on_poll(poll_res)

    def _on_sync_complete(self) -> None:
        self._publish_fn = self._store_and_broadcast
        self._keep_samples()

    def _poll_data_packets(self) -> None:
        """Receive data packets in a steady state.

        Gets called every time one of the requestes modalities produced new data.
        In normal operation mode, all messages are 2-part.
        """
        topic, payload = self._sub.recv_multipart()
        receive_time = get_time()
        msg = deserialize(payload)
        topic_tree: list[str] = topic.decode("utf-8").split(".")
        self._in_streams[topic_tree[0]].append_data(process_time_s=receive_time, **msg)
        self._process_data(topic=topic_tree[0], msg=msg)

    def _poll_ending_data_packets(self) -> None:
        """Receive data packets from producers and monitor for end-of-stream signal.

        When system triggered a safe exit, Pipeline gets a mix of normal 2-part messages
        and 3-part 'END' message from each Producer that safely exited.
        It's more efficient to dynamically switch the callback instead of checking every message.

        Processes packets on each modality until all data sources sent the 'END' packet.
        If triggered to stop and no more available data, sends empty 'END' packet and joins.
        """
        topic, payload = self._sub.recv_multipart()
        receive_time = get_time()
        # 'END' empty packet from a Producer.
        if CMD_END.encode("utf-8") in payload:
            topic_tree: list[str] = topic.decode("utf-8").split(".")
            self._is_producer_ended[topic_tree[0]] = True
            if all(list(self._is_producer_ended.values())):
                self._is_more_data_in = False
                self._check_before_send_end_packet()
        # Regular data packets.
        else:
            msg = deserialize(payload)
            topic_tree: list[str] = topic.decode("utf-8").split(".")
            self._in_streams[topic_tree[0]].append_data(
                process_time_s=receive_time, **msg
            )
            self._process_data(topic=topic_tree[0], msg=msg)

    def _publish(self, tag: str, **kwargs) -> None:
        """Pass generated data to the ZeroMQ message exchange layer.

        Args:
            tag (str): Uniquely identifying key for the data generated by the Node.
        """
        self._publish_fn(tag, **kwargs)

    def _store_and_broadcast(self, tag: str, process_time_s: float, **kwargs) -> None:
        """Place captured data into the corresponding Stream datastructure and transmit serialized ZeroMQ packets to subscribers.

        Args:
            tag (str): Uniquely identifying key for the modality to label data for message exchange.
        """
        msg = serialize(**kwargs)
        self._pub.send_multipart([tag.encode("utf-8"), msg])
        self._out_stream.append_data(process_time_s=process_time_s, **kwargs)

    def _trigger_stop(self):
        self._poll_data_fn = self._poll_ending_data_packets
        self._stop_new_data()

    def _notify_no_more_data_out(self) -> None:
        """Notify the Pipeline that no more data will be generated."""
        self._is_more_data_out = False
        self._poller.unregister(self._pub)
        self._check_before_send_end_packet()

    def _check_before_send_end_packet(self) -> None:
        """Check whether both input and output data streams have ended, and send 'END' packet if so."""
        if not self._is_more_data_in and (not self._is_async_generate or (self._is_async_generate and not self._is_more_data_out)):
            self._send_end_packet()

    def _send_end_packet(self) -> None:
        """Send 'END' empty packet and label Node as done to safely finish and exit the process and its threads."""
        self._pub.send_multipart(
            [
                ("%s.data" % self._log_source_tag()).encode("utf-8"),
                b"",
                CMD_END.encode("utf-8"),
            ]
        )
        self._is_done = True

    @abstractmethod
    def _cleanup(self) -> None:
        self._storage.cleanup()
        # Before closing the PUB socket, wait for the 'BYE' signal from the Broker.
        self._sync.send_multipart(
            [self._log_source_tag().encode("utf-8"), CMD_EXIT.encode("utf-8")]
        )
        host, cmd = (
            self._sync.recv_multipart()
        )  # no need to read contents of the message.
        print(
            "%s received %s from %s."
            % (self._log_source_tag(), cmd.decode("utf-8"), host.decode("utf-8")),
            flush=True,
        )
        self._pub.close()
        self._sub.close()
        # Join on the logging background thread last, so that all things can finish in parallel.
        self._storage_thread.join()
        super()._cleanup()



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

from abc import abstractmethod
import asyncio
from collections import OrderedDict

from hermes.base.state_interface import StateInterface
from hermes.base.stream import Stream
from hermes.base.storage.storage_interface import StorageInterface


class AbstractStorageState(StateInterface):
    """Abstract class for the Storage component."""

    def __init__(self, context: StorageInterface):
        """Constructor of the AbstractStorageState.

        Args:
            context (StorageInterface): Reference to the Storage object.
        """
        self._context = context
        self._is_continue_fsm = True

    @abstractmethod
    def run(self) -> None:
        pass

    def is_continue(self) -> bool:
        return self._is_continue_fsm

    def kill(self) -> None:
        pass


class StartState(AbstractStorageState):
    """Startup state that initializes Storage.

    Will immediately transition into `StreamState` after initialization.
    """

    def __init__(self, context: StorageInterface, streams: OrderedDict[str, Stream]):
        """Constructor of the StartState.

        Args:
            context (StorageInterface): Reference to the Storage object.
            streams (OrderedDict[str, Stream]): Reference to the mapping between uniquly identifying Node keys and the corresponding Stream datastructures.
        """
        super().__init__(context)
        self._context._initialize(streams)

    def run(self) -> None:
        self._context._set_state(StreamState(self._context))


class StreamState(AbstractStorageState):
    """Streaming state of the Storage component.

    Will periodically flush data to disk, to clear memory,
    if any streams were specified to stream.

    Will flush remaining data to disk on exit.

    Using some streams in stream and others in dump has undefined behavior.
    """

    def __init__(self, context: StorageInterface):
        """Constructor of the StreamState.

        Args:
            context (StorageInterface): Reference to the Storage object.
        """
        super().__init__(context)
        # Prepare stream-logging.
        if self._context._is_to_stream():
            self._context._start_stream_logging()

    def run(self) -> None:
        asyncio.run(self._context._log_data())
        self._context._release_thread_pool()
        self._is_continue_fsm = False
        # self._context._set_state(DumpState(self._context))


class DumpState(AbstractStorageState):
    """Passive recording state of the Storage component.

    Will flush data to disk once upon system termination.

    May run out of memory if the recording is long or
    user doesn't provision sufficient data.

    Using some streams in stream and others in dump has undefined behavior.
    """

    def __init__(self, context: StorageInterface):
        """Constructor of the DumpState.

        Args:
            context (StorageInterface): Reference to the Storage object.
        """
        super().__init__(context)
        # Dump write files at the end of the trial for data that hadn't been streamed.
        #   Assumes all intermediate recorded data can fit in memory.
        if self._context._is_to_dump():
            self._context._start_dump_logging()

    def run(self) -> None:
        # Until top-level module's main thread indicated that it finished producing data,
        #   periodically sleep the thread to yield the CPU.
        self._context._wait_till_flush()
        # TODO: When experiment ended, write data once and wrap up.
        #   Use the end log time to not overwrite written streamed data with empty dumped files?
        asyncio.run(self._context._log_data())
        # Release Thread Pool used for AsyncIO for file writing.
        self._context._release_thread_pool()
        self._is_continue_fsm = False



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

import time as _time
from time import perf_counter
from threading import Lock
from datetime import datetime, timezone
from dateutil import tz


class SingletonMeta(type):
    _instances = {}
    _lock: Lock = Lock()

    def __call__(cls, *args, **kwargs):
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]


class SystemTime(metaclass=SingletonMeta):
    """Highly accurate system time Singleton that uses performance counters of the host device.

    Performance counters of the host devices are not Unix float numbers since epoch,
    but a float starting from 0 on the boot-up of the device. This Singleton uses less accurate system time
    with the `time.time()` call and combines with the performance counters, to get highly accurate time of the host device.

    This Singleton's methods should not be called directly. Use the global methods instead.

    Each subprocess of the application must be initialized via `init_time(...)`
    with a reference time obtained in the parent process via `get_ref_time()`.
    """

    def __init__(self):
        rough_time = _time.time()
        counter = perf_counter()
        self._ref_time = rough_time - counter
        # print(
        #     f"Reference time: {self._ref_time}, rough: {rough_time}, counter: {counter}",
        #     flush=True,
        # )

    def time(self) -> float:
        return self._ref_time + perf_counter()

    def _get_ref_time(self) -> float:
        return self._ref_time

    def _set_ref_time(self, ref_time: float) -> None:
        self._ref_time = ref_time


def init_time(ref_time: float) -> None:
    """Initialize the current process's `SystemTime` Singleton with a common reference time.

    Args:
        ref_time (float): Time obtained via `get_ref_time()` in the parent process to use as a reference for performance counters in the current process.
    """
    SystemTime()._set_ref_time(ref_time)
    # print(
    #     f"Initialized system time with ref time: {ref_time} at {get_time()}", flush=True
    # )


def get_ref_time() -> float:
    """Gets the current process's `SystemTime` Singleton reference time."""
    return SystemTime()._get_ref_time()


def get_time() -> float:
    """Gets the highly accurate current system time."""
    return SystemTime().time()


def get_time_str(time_s: float = get_time(), format: str = "%Y-%m-%d_%H-%M-%S") -> str:
    """Gets a date string from seconds since epoch.

    Args:
        time_s (float, optional): Time since epoch to convert to the human-readable string. Defaults to the current system time.
        format (str, optional): Format to construct the date string. Default to `%Y-%m-%d_%H-%M-%S`.
    
    Returns:
        str: Formatted date string corresponding to the provided time since epoch. 
    """
    time_datetime = datetime.fromtimestamp(time_s)
    time_str = time_datetime.strftime(format)
    return time_str


def get_time_s_from_utc_time_no_date_str(
    time_utc_str: str,
    input_time_format: str = "%H:%M:%S.%f",
    date_utc_str: str | None = None,
    input_date_format: str = "%Y-%m-%d",
) -> float:
    """Gets local time in seconds since epoch from provided UTC time strings.
    
    Args:
        time_utc_str (str): UTC time string to convert to seconds.
        input_time_format (str, optional): Time string format of the provided time. Defaults to `%H:%M:%S.%f`.
        date_utc_str (str, optional): UTC date to convert to seconds. Defaults to `None`.
        input_date_format (str, optional): Date string format of the provided date. Defaults to `%Y-%m-%d`.

    Returns:
        float: Seconds count since epoch, corresponding to the provided UTC time strings.
    """
    from_zone = tz.tzutc()
    to_zone = tz.tzlocal()

    # Get the current UTC date if no date was provided.
    if date_utc_str is None:
        now_utc_datetime = datetime.now(timezone.utc)
        date_utc_str = now_utc_datetime.strftime(input_date_format)

    # Combine the date and time.
    utc_str = "%s %s" % (date_utc_str, time_utc_str)
    utc_datetime = datetime.strptime(
        utc_str, input_date_format + " " + input_time_format
    )

    # Convert to local time, then to seconds since epoch.
    utc_datetime = utc_datetime.replace(tzinfo=from_zone)
    local_datetime = utc_datetime.astimezone(to_zone)
    return local_datetime.timestamp()


def get_time_s_from_local_str(
    time_local_str: str,
    input_time_format: str = "%H:%M:%S.%f",
    date_local_str: str | None = None,
    input_date_format: str = "%Y-%m-%d",
) -> float:
    """Gets seconds since epoch from provided local time string.
    
    Args:
        time_local_str (str): local time string to convert to seconds.
        input_time_format (str, optional): Time string format of the provided time. Defaults to `%H:%M:%S.%f`.
        date_local_str (str, optional): local date string to convert to seconds. Defaults to `None`.
        input_date_format (str, optional): Date string format of the provided date. Defaults to `%Y-%m-%d`.

    Returns:
        float: Seconds count since epoch, corresponding to the provided local time strings.
    """
    # Get the current local date if no date was provided.
    if date_local_str is None:
        now_local_datetime = datetime.now()
        date_local_str = now_local_datetime.strftime(input_date_format)

    # Combine the date and time.
    local_str = "%s %s" % (date_local_str, time_local_str)
    local_datetime = datetime.strptime(
        local_str, input_date_format + " " + input_time_format
    )

    return local_datetime.timestamp()



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

import queue
from typing import Dict, Any, Callable
from dataclasses import dataclass
from collections import defaultdict
import threading

from hermes.utils.time_utils import get_time


@dataclass
class DataRequest:
    """Object wrapping user's element of interest into a fetch request."""

    key: str
    timestamp: float


class Cache:
    """Caching module prefetches segments of data that has long IO, for more responsive experience."""

    def __init__(self, fetch_fn: Callable[[Any], Dict[Any, Any]], fetch_offset: int):
        """Constructor of the Cache component for long IO prefetching to mask latency behind wide bandwidth.

        Args:
            fetch_fn (Callable[[Any], Dict[Any, Any]]): User-provided function with long IO to use for asynchronous prefetching (database operation, API call, FFmpeg decoding, etc.).
            fetch_offset (int): How many elements before the requested to cache in case of jumping backwards during playback.
        """
        self._cache: Dict[Any, Any] = {}
        self._fetch_fn = fetch_fn
        self._fetch_offset = fetch_offset
        self._request_queue: queue.Queue[DataRequest] = queue.Queue()
        self._data_events: Dict[Any, threading.Event] = defaultdict(threading.Event)

    def start(self):
        """Start the background cache management thread."""
        self._cache_task = threading.Thread(target=self._run_cache_manager)
        self._cache_task.start()

    def join(self):
        """Wait on the background cache management thread."""
        if hasattr(self, "cache_task"):
            self._cache_task.join()

    def get_data(self, key: Any) -> Any:
        """Request data from the cache.

        Checks if data is already in the cache.
        Adds a request to the queue for background processing.
        Returns immediately if requested data is cached, otherwise waits for the background task to fetch it.
        Stalls until the data is available in the cache for the given key.

        Args:
            key (Any): Unique key correctly identifying the element of interest in the user-provided fetch function.

        Returns:
            Any: The requested element of interest.
        """
        if key in self._cache:
            return self._cache[key]

        request = DataRequest(key, get_time())
        self._request_queue.put(request)

        event = self._data_events[key]
        event.wait()
        del self._data_events[key]
        return self._cache[key]

    def _run_cache_manager(self):
        """Main loop of the background continuous cache management thread.

        Processes requests from the user and prefetches most likely used next segment of elements.
        """
        while True:
            try:
                request = self._request_queue.get(timeout=None)
                self._process_request(request)
            except queue.Empty:
                print(f"Timeout: no new cache fill request")
            except Exception as e:
                print(f"Error in cache manager: {e}")

    def _process_request(self, request: DataRequest):
        """Process user request for the next element of interest.

        Runs user-provided long-IO fetch procedure if the data is not immediately available in cache or isn't already being fetched.
        Updates the cache and notifies all watchers waiting for this element.

        Args:
            request (DataRequest): User-requested element of interest.
        """
        if request.key not in self._cache:
            self._cache = self._fetch(request.key)
            if request.key in self._data_events:
                self._data_events[request.key].set()

    def _fetch(self, key: Any) -> Dict[Any, Any]:
        """Fetches data from an external source using user-defined IO function.

        Returns:
            Dict[Any, Any]: Mapping from a unique key to the element of interest.
        """
        if (window_start_frame := key - self._fetch_offset) > 0:
            return self._fetch_fn(window_start_frame)
        else:
            return self._fetch_fn(0)



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

from argparse import Namespace
from multiprocessing import Queue
from multiprocessing.synchronize import Event as _Event
from typing import Callable

from hermes.base.broker.broker import Broker


def launch_broker(
    args: Namespace,
    node_specs: list[dict],
    input_queue: "Queue[tuple[float, str]]",
    is_ready_event: _Event,
    is_quit_event: _Event,
    is_done_event: _Event,
    ref_time_s: float,
) -> None:
    """Builds the `Broker` using provided configurations and manage all the components of the experiment.

    Meant to be used as a target for a spawned Process or Thread.

    Args:
        args (argparse.Namespace): HERMES top-level user input arguments for the experiment definition.
        node_specs (list[dict]): Configuration dictionaries of the Broker's locally managed `Node`s.
        input_queue (Queue[tuple[float, str]]): Multiprocessing queue to fan-out user keyboard inputs to all `Broker`s locally managed `Node`s.
        is_ready_event (Event): Multiprocessing synchronization primitive indicating completion of `Broker`s setup and handshaking with `Node`s.
        is_quit_event (Event): Multiprocessing synchronization primitive triggering to the `Broker` to gracefully wrap up and end.
        is_done_event (Event): Multiprocessing synchronization primitive indicating that the `Broker` is finished and experiment ended.
        ref_time_s (float): Main process reference system time obtained with `get_ref_time()` to use in all child processes for syncing the data.
    """
    local_broker: Broker = Broker(
        host_ip=args.host_ip,
        node_specs=node_specs,
        is_ready_event=is_ready_event,
        is_quit_event=is_quit_event,
        is_done_event=is_done_event,
        input_queue=input_queue,
        is_master_broker=args.is_master_broker,
        ref_time_s=ref_time_s,
    )

    # Connect broker to remote publishers at the wearable PC to get data from the wearable sensors.
    for ip in args.remote_publisher_ips:
        local_broker.connect_to_remote_broker(addr=ip)

    # Expose local wearable data to remote subscribers (e.g. edge server).
    if args.remote_subscriber_ips:
        local_broker.expose_to_remote_broker(args.remote_subscriber_ips)

    # Subscribe to the KILL signal of a remote machine.
    if args.is_remote_kill:
        local_broker.subscribe_to_killsig(addr=args.remote_kill_ip)

    # Only master host runs with duration, others wait for commands.
    if args.is_master_broker:
        local_broker(args.duration_s)
    else:
        local_broker()


def launch_callable(obj: Callable, *args, **kwargs) -> None:
    """Launches a callable object with the user-provided inputs.

    Args:
        args (list): Ordered unnamed inputs to provide to the callable object.
        kwargs (dict): Named key-value inputs to provide to the callable object.
    """
    obj(*args, **kwargs)


def launch_handler(module: type, *args, **kwargs) -> None:
    """Used as a target to instantiate a runnable object as a subprocess.

    Args:
        args (list): Ordered unnamed arguments to instantiate the object with.
        kwargs (dict): Named key-value arguments to instantiate the object with.
    """
    obj = module(*args, **kwargs)
    obj()



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

import importlib
from typing import Any


def search_module_class(module_name: str, class_name: str) -> type[Any]:
    """Queries the current Python environment to match the requested `hermes.<module>`.

    Args:
        module_name (str): Name of the Python module containing the requested HERMES supported class.
        class_name (str): Name of the class in the provided module to retrieve for construction.
    """
    module_path = "hermes.%s" % module_name

    try:
        module = importlib.import_module(module_path)
    except ImportError as e:
        print(e, flush=True)
        raise ImportError(
            "Could not import subpackage '%s'. "
            "Ensure it is installed: pip install pysio-hermes-%s"
            % (module_name, module_name)
        ) from e

    if not hasattr(module, class_name):
        raise AttributeError(
            "Class '%s' not found in module '%s'. "
            "Check the spelling of the class name." % (class_name, module_name)
        )

    class_type = getattr(module, class_name)

    return class_type



############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

from abc import abstractmethod

from hermes.base.nodes.node_interface import NodeInterface
from hermes.base.stream import Stream


class PipelineInterface(NodeInterface):
    """Interface for the Pipeline Node component."""

    @classmethod
    @abstractmethod
    def create_stream(cls, stream_spec: dict) -> Stream:
        """Instantiate Stream datastructure object specific to this Pipeline.

        Should also be a class method to create Stream objects on consumers.

        Args:
            stream_spec (dict): Mapping of corresponding Stream object parameters to user-defined configuration values.

        Returns:
            Stream: Datastructure object of the corresponding Node, configured according to the user-provided specification.
        """
        pass

    @abstractmethod
    def _process_data(self, topic: str, msg: dict) -> None:
        """Main iteration loop logic for the Node during its running phase.

        Contained logic has to deal with async multiple modalities.
        Must end with calling `_send_end_packet`.

        Args:
            topic (str): Uniquely identified modality of the contained data.
            msg (dict): Received data of the corresponding modality.
        """
        pass

    @abstractmethod
    def _keep_samples(self) -> None:
        """Node-specific externally triggered function to start keeping in memory streamed data."""
        pass

    @abstractmethod
    def _generate_data(self) -> None:
        """Main iteration loop logic to process and distribute internal asynchronously generated data.
        
        Contained logic must deal with sending internally generated data packets until external termination
        signal is received, in a non-deadlocking way to the rest of the Pipeline processing.
        """
        pass

    @abstractmethod
    def _stop_new_data(self) -> None:
        """Stop sampling data, continue sending already captured until none is left."""
        pass
